import pytest
from git.exc import HookExecutionError
from nbformat.v4.nbbase import new_code_cell, new_markdown_cell, new_notebook
from pre_commit.main import main as pre_commit

from jupytext import read, write

from .utils import (
    requires_pandoc,
    skip_pre_commit_tests_on_windows,
    skip_pre_commit_tests_when_jupytext_folder_is_not_a_git_repo,
)


@requires_pandoc
@skip_pre_commit_tests_on_windows
@skip_pre_commit_tests_when_jupytext_folder_is_not_a_git_repo
def test_pre_commit_hook_sync_reformat_code_and_markdown(
    tmpdir,
    cwd_tmpdir,
    tmp_repo,
    jupytext_repo_root,
    jupytext_repo_rev,
    notebook_with_outputs,
):
    """Here we sync the ipynb notebook with a py:percent file and also apply black and pandoc to reformat both
    code and markdown cells.

    Note: the new cell ids introduced in nbformat 5.1.0 are not yet supported by all the programs that treat
    ipynb files. Consequently we pin the version of nbformat to 5.0.8 in all the environments below (and you
    will have to do the same on the environment in which you edit the notebooks).
    """
    pre_commit_config_yaml = f"""
repos:
- repo: {jupytext_repo_root}
  rev: {jupytext_repo_rev}
  hooks:
  - id: jupytext
    args: [--sync, --pipe-fmt, ipynb, --pipe, 'pandoc --from ipynb --to ipynb --markdown-headings=atx', --show-changes]
    additional_dependencies:
    - nbformat==5.0.8  # because pandoc 2.11.4 does not preserve yet the new cell ids
  - id: jupytext
    args: [--sync, --pipe, black, --show-changes]
    additional_dependencies:
    - black==20.8b1  # Matches black hook below
    - nbformat==5.0.8  # for compatibility with the pandoc hook above

- repo: https://github.com/psf/black
  rev: 20.8b1
  hooks:
  - id: black
"""
    tmpdir.join(".pre-commit-config.yaml").write(pre_commit_config_yaml)

    tmp_repo.git.add(".pre-commit-config.yaml")
    pre_commit(["install", "--install-hooks", "-f"])

    tmpdir.join(".jupytext.toml").write('formats = "ipynb,py:percent"')
    tmp_repo.git.add(".jupytext.toml")
    tmp_repo.index.commit("pair notebooks")

    # write a test notebook
    notebook = new_notebook(
        cells=[
            new_code_cell("1+1"),
            new_markdown_cell(
                """This is a complex markdown cell

# With a h1 header
## And a h2 header

| And       | A  | Table |
| --------- | ---| ----- |
| 0         | 1  | 2     |

!!!WARNING!!! This hook does not seem compatible with
explicit paragraph breaks (two spaces at the end of a line).

And a VERY long line.
""".replace(
                    "VERY ", "very " * 51
                )
            ),
        ],
        metadata=notebook_with_outputs.metadata,
    )

    # We write the notebook in version 4.4, i.e. with the equivalent of nbformat version 5.0.8
    notebook.nbformat = 4
    notebook.nbformat_minor = 4
    write(notebook, "test.ipynb")

    # try to commit it, should fail because
    # 1. the py version hasn't been added
    # 2. the first cell is '1+1' which is not black compliant
    # 3. the second cell needs to be wrapped
    tmp_repo.git.add("test.ipynb")
    with pytest.raises(
        HookExecutionError,
        match="files were modified by this hook",
    ):
        tmp_repo.index.commit("failing")

    # Add the two files
    tmp_repo.git.add("test.ipynb")
    tmp_repo.git.add("test.py")

    # now the commit will succeed
    tmp_repo.index.commit("passing")
    assert "test.ipynb" in tmp_repo.tree()
    assert "test.py" in tmp_repo.tree()

    # both the code and the markdown cells were reformatted
    nb = read("test.ipynb")
    assert nb.cells[0].source == "1 + 1"

    print(nb.cells[1].source)
    assert (
        nb.cells[1].source
        == """This is a complex markdown cell

# With a h1 header

## And a h2 header

| And | A   | Table |
|-----|-----|-------|
| 0   | 1   | 2     |

!!!WARNING!!! This hook does not seem compatible with explicit paragraph
breaks (two spaces at the end of a line).

And a very very very very very very very very very very very very very
very very very very very very very very very very very very very very
very very very very very very very very very very very very very very
very very very very very very very very very very long line."""
    )



import json
import os
import re
import sys
from pathlib import Path

import pytest
from jupyter_client.kernelspec import find_kernel_specs, get_kernel_spec

from jupytext.cell_reader import rst2md
from jupytext.cli import system
from jupytext.myst import is_myst_available
from jupytext.pandoc import is_pandoc_available
from jupytext.quarto import is_quarto_available


def tool_version(tool):
    try:
        args = tool.split(" ")
        args.append("--version")
        return system(*args)
    except (OSError, SystemExit):  # pragma: no cover
        return None


def isort_version():
    try:
        import isort

        return isort.__version__
    except ImportError:
        return None


requires_jupytext_installed = pytest.mark.skipif(
    not tool_version("jupytext"), reason="jupytext is not installed"
)
requires_black = pytest.mark.skipif(not tool_version("black"), reason="black not found")
requires_isort = pytest.mark.skipif(
    not isort_version() or isort_version() <= "5.3.0",
    reason="isort not found",
)
requires_flake8 = pytest.mark.skipif(
    not tool_version("flake8"), reason="flake8 not found"
)
requires_autopep8 = pytest.mark.skipif(
    not tool_version("autopep8"), reason="autopep8 not found"
)
requires_nbconvert = pytest.mark.skipif(
    not tool_version("jupyter nbconvert"), reason="nbconvert not found"
)
requires_sphinx_gallery = pytest.mark.skipif(
    not rst2md, reason="sphinx_gallery is not available"
)
requires_pandoc = pytest.mark.skipif(
    # The mirror files changed slightly when Pandoc 2.11 was introduced
    # https://github.com/mwouts/jupytext/commit/c07d919702999056ce47f92b74f63a15c8361c5d
    not is_pandoc_available(min_version="2.11"),
    reason="pandoc>=2.11 is not available",
)
requires_quarto = pytest.mark.skipif(
    not is_quarto_available(min_version="0.2.0"), reason="quarto>=0.2 is not available"
)
requires_no_pandoc = pytest.mark.skipif(
    is_pandoc_available(), reason="Pandoc is installed"
)
requires_ir_kernel = pytest.mark.skipif(
    not any(get_kernel_spec(name).language == "R" for name in find_kernel_specs()),
    reason="irkernel is not installed",
)
requires_user_kernel_python3 = pytest.mark.skipif(
    "python_kernel" not in find_kernel_specs(),
    reason="Please run 'python -m ipykernel install --name python_kernel --user'",
)
requires_myst = pytest.mark.skipif(
    not is_myst_available(), reason="myst_parser not found"
)
requires_no_myst = pytest.mark.skipif(is_myst_available(), reason="myst is available")
skip_on_windows = pytest.mark.skipif(sys.platform.startswith("win"), reason="Issue 489")
skip_pre_commit_tests_on_windows = pytest.mark.skipif(
    sys.platform.startswith("win"),
    reason="OSError: [WinError 193] %1 is not a valid Win32 application",
)
skip_pre_commit_tests_when_jupytext_folder_is_not_a_git_repo = pytest.mark.skipif(
    not (Path(__file__).parent.parent / ".git").is_dir(),
    reason="Jupytext folder is not a git repository #814",
)


def list_notebooks(path="ipynb", skip="World"):
    """All notebooks in the directory notebooks/path,
    or in the package itself"""
    if path == "ipynb":
        return (
            list_notebooks("ipynb_julia", skip=skip)
            + list_notebooks("ipynb_py", skip=skip)
            + list_notebooks("ipynb_R", skip=skip)
        )

    if path == "ipynb_all":
        all_notebooks = []

        nb_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "notebooks")
        for dir in os.listdir(nb_path):
            if dir.startswith("ipynb_"):
                all_notebooks.extend(list_notebooks(dir, skip=skip))

        return all_notebooks

    nb_path = os.path.dirname(os.path.abspath(__file__))
    if path.startswith("."):
        nb_path = os.path.join(nb_path, path)
    else:
        nb_path = os.path.join(nb_path, "notebooks", path)

    if skip:
        skip_re = re.compile(".*" + skip + ".*")
        notebooks = [
            os.path.join(nb_path, nb_file)
            for nb_file in os.listdir(nb_path)
            if not skip_re.match(nb_file)
        ]
    else:
        notebooks = [os.path.join(nb_path, nb_file) for nb_file in os.listdir(nb_path)]

    # ignore ".ipynb_checkpoints" sub-folder
    return [nb_file for nb_file in notebooks if os.path.isfile(nb_file)]


def notebook_model(nb):
    """Return a notebook model, with content a dictionary rather than a notebook object"""
    return dict(type="notebook", content=json.loads(json.dumps(nb)))



import pytest

from jupytext import read, reads, writes
from jupytext.formats import JupytextFormatError, auto_ext_from_metadata

from .utils import list_notebooks


@pytest.mark.parametrize(
    "nb_file", list_notebooks("ipynb_R") + list_notebooks("ipynb_py")
)
def test_auto_in_fmt(nb_file):
    nb = read(nb_file)
    auto_ext = auto_ext_from_metadata(nb.metadata)
    fmt = auto_ext[1:] + ":percent"
    text = writes(nb, "auto:percent")

    assert "auto" not in text

    nb2 = reads(text, fmt)
    assert nb2.metadata["jupytext"]["text_representation"]["extension"]
    assert nb2.metadata["jupytext"]["text_representation"]["format_name"] == "percent"

    del nb.metadata["language_info"]
    del nb.metadata["kernelspec"]
    with pytest.raises(JupytextFormatError):
        writes(nb, "auto:percent")


@pytest.mark.parametrize("nb_file", list_notebooks("ipynb_all"))
def test_auto_from_kernelspecs_works(nb_file):
    nb = read(nb_file)
    language_info = nb.metadata.pop("language_info")
    expected_ext = language_info.get("file_extension")
    if not expected_ext:
        pytest.skip("No file_extension in language_info")
    if expected_ext == ".r":
        expected_ext = ".R"
    elif expected_ext == ".fs":
        expected_ext = ".fsx"
    auto_ext = auto_ext_from_metadata(nb.metadata)
    if auto_ext == ".sage":
        pytest.xfail(
            "Sage notebooks have Python in their language_info metadata, see #727"
        )
    assert auto_ext == expected_ext


@pytest.mark.parametrize(
    "nb_file",
    list_notebooks("ipynb_R") + list_notebooks("ipynb_py", skip="(World|plotly)"),
)
def test_auto_in_formats(nb_file):
    nb = read(nb_file)
    nb.metadata["jupytext"] = {"formats": "ipynb,auto:percent"}
    fmt = auto_ext_from_metadata(nb.metadata)[1:] + ":percent"
    expected_formats = "ipynb," + fmt

    text = writes(nb, "ipynb")
    assert "auto" not in text
    nb2 = reads(text, "ipynb")
    assert nb2.metadata["jupytext"]["formats"] == expected_formats

    text = writes(nb, "auto:percent")
    assert "auto" not in text
    nb2 = reads(text, fmt)
    assert nb2.metadata["jupytext"]["formats"] == expected_formats

    del nb.metadata["language_info"]
    del nb.metadata["kernelspec"]
    with pytest.raises(JupytextFormatError):
        writes(nb, "ipynb")
    with pytest.raises(JupytextFormatError):
        writes(nb, "auto:percent")



"""Compare two Jupyter notebooks"""

import difflib
import json
import re

from .cell_metadata import _IGNORE_CELL_METADATA
from .combine import combine_inputs_with_outputs
from .formats import long_form_one_format
from .header import _DEFAULT_NOTEBOOK_METADATA
from .jupytext import reads, writes
from .metadata_filter import filter_metadata

_BLANK_LINE = re.compile(r"^\s*$")


def _multilines(obj):
    try:
        lines = obj.splitlines()
        return lines + [""] if obj.endswith("\n") else lines
    except AttributeError:
        # Remove the final blank space on Python 2.7
        # return json.dumps(obj, indent=True, sort_keys=True).splitlines()
        return [
            line.rstrip()
            for line in json.dumps(obj, indent=True, sort_keys=True).splitlines()
        ]


def compare(
    actual, expected, actual_name="actual", expected_name="expected", return_diff=False
):
    """Compare two strings, lists or dict-like objects"""
    if actual != expected:
        diff = difflib.unified_diff(
            _multilines(expected),
            _multilines(actual),
            expected_name,
            actual_name,
            lineterm="",
        )
        if expected_name == "" and actual_name == "":
            diff = list(diff)[2:]
        diff = "\n".join(diff)
        if return_diff:
            return diff
        raise AssertionError("\n" + diff)
    return "" if return_diff else None


def filtered_cell(cell, preserve_outputs, cell_metadata_filter):
    """Cell type, metadata and source from given cell"""
    filtered = {
        "cell_type": cell.cell_type,
        "source": cell.source,
        "metadata": filter_metadata(
            cell.metadata, cell_metadata_filter, _IGNORE_CELL_METADATA
        ),
    }

    if preserve_outputs:
        for key in ["execution_count", "outputs"]:
            if key in cell:
                filtered[key] = cell[key]

    return filtered


def filtered_notebook_metadata(notebook, ignore_display_name=False):
    """Notebook metadata, filtered for metadata added by Jupytext itself"""
    metadata = filter_metadata(
        notebook.metadata,
        notebook.metadata.get("jupytext", {}).get("notebook_metadata_filter"),
        _DEFAULT_NOTEBOOK_METADATA,
    )

    # The display name for the kernel might change (Quarto format on the CI)
    if ignore_display_name:
        metadata.get("kernelspec", {}).pop("display_name", None)

    if "jupytext" in metadata:
        del metadata["jupytext"]
    return metadata


class NotebookDifference(Exception):
    """Report notebook differences"""


def same_content(ref_source, test_source, allow_removed_final_blank_line):
    """Is the content of two cells the same, except for an optional final blank line?"""
    if ref_source == test_source:
        return True

    if not allow_removed_final_blank_line:
        return False

    # Is ref identical to test, plus one blank line?
    ref_source = ref_source.splitlines()
    test_source = test_source.splitlines()

    if not ref_source:
        return False

    if ref_source[:-1] != test_source:
        return False

    return _BLANK_LINE.match(ref_source[-1])


def compare_notebooks(
    notebook_actual,
    notebook_expected,
    fmt=None,
    allow_expected_differences=True,
    raise_on_first_difference=True,
    compare_outputs=False,
    compare_ids=None,
):
    """Compare the two notebooks, and raise with a meaningful message
    that explains the differences, if any"""
    fmt = long_form_one_format(fmt)
    format_name = fmt.get("format_name")

    if (
        format_name == "sphinx"
        and notebook_actual.cells
        and notebook_actual.cells[0].source == "%matplotlib inline"
    ):
        notebook_actual.cells = notebook_actual.cells[1:]

    if compare_ids is None:
        compare_ids = compare_outputs

    modified_cells, modified_cell_metadata = compare_cells(
        notebook_actual.cells,
        notebook_expected.cells,
        raise_on_first_difference,
        compare_outputs=compare_outputs,
        compare_ids=compare_ids,
        cell_metadata_filter=notebook_actual.get("jupytext", {}).get(
            "cell_metadata_filter"
        ),
        allow_missing_code_cell_metadata=(
            allow_expected_differences and format_name == "sphinx"
        ),
        allow_missing_markdown_cell_metadata=(
            allow_expected_differences and format_name in ["sphinx", "spin"]
        ),
        allow_filtered_cell_metadata=allow_expected_differences,
        allow_removed_final_blank_line=allow_expected_differences,
    )

    # Compare notebook metadata
    modified_metadata = False
    try:
        ignore_display_name = (
            fmt.get("extension") == ".qmd" and allow_expected_differences
        )
        compare(
            filtered_notebook_metadata(notebook_actual, ignore_display_name),
            filtered_notebook_metadata(notebook_expected, ignore_display_name),
        )
    except AssertionError as error:
        if raise_on_first_difference:
            raise NotebookDifference("Notebook metadata differ: {}".format(str(error)))
        modified_metadata = True

    error = []
    if modified_cells:
        error.append(
            "Cells {} differ ({}/{})".format(
                ",".join([str(i) for i in modified_cells]),
                len(modified_cells),
                len(notebook_expected.cells),
            )
        )
    if modified_cell_metadata:
        error.append(
            "Cell metadata '{}' differ".format(
                "', '".join([str(i) for i in modified_cell_metadata])
            )
        )
    if modified_metadata:
        error.append("Notebook metadata differ")

    if error:
        raise NotebookDifference(" | ".join(error))


def compare_cells(
    actual_cells,
    expected_cells,
    raise_on_first_difference=True,
    compare_outputs=True,
    compare_ids=True,
    cell_metadata_filter=None,
    allow_missing_code_cell_metadata=False,
    allow_missing_markdown_cell_metadata=False,
    allow_filtered_cell_metadata=False,
    allow_removed_final_blank_line=False,
):
    """Compare two collection of notebook cells"""
    test_cell_iter = iter(actual_cells)
    modified_cells = set()
    modified_cell_metadata = set()

    for i, ref_cell in enumerate(expected_cells, 1):
        try:
            test_cell = next(test_cell_iter)
        except StopIteration:
            if raise_on_first_difference:
                raise NotebookDifference(
                    "No cell corresponding to {} cell #{}:\n{}".format(
                        ref_cell.cell_type, i, ref_cell.source
                    )
                )
            modified_cells.update(range(i, len(expected_cells) + 1))
            break

        ref_lines = [
            line for line in ref_cell.source.splitlines() if not _BLANK_LINE.match(line)
        ]
        test_lines = []

        # 1. test cell type
        if ref_cell.cell_type != test_cell.cell_type:
            if raise_on_first_difference:
                raise NotebookDifference(
                    "Unexpected cell type '{}' for {} cell #{}:\n{}".format(
                        test_cell.cell_type, ref_cell.cell_type, i, ref_cell.source
                    )
                )
            modified_cells.add(i)

        # Compare cell ids (introduced in nbformat 5.1.0)
        if compare_ids and test_cell.get("id") != ref_cell.get("id"):
            if raise_on_first_difference:
                raise NotebookDifference(
                    f"Cell ids differ on {test_cell['cell_type']} cell #{i}: "
                    f"'{test_cell.get('id')}' != '{ref_cell.get('id')}'"
                )
            modified_cells.add(i)

        # 2. test cell metadata
        if (ref_cell.cell_type == "code" and not allow_missing_code_cell_metadata) or (
            ref_cell.cell_type != "code" and not allow_missing_markdown_cell_metadata
        ):

            ref_metadata = ref_cell.metadata
            test_metadata = test_cell.metadata
            if allow_filtered_cell_metadata:
                ref_metadata = {
                    key: ref_metadata[key]
                    for key in ref_metadata
                    if key not in _IGNORE_CELL_METADATA
                }
                test_metadata = {
                    key: test_metadata[key]
                    for key in test_metadata
                    if key not in _IGNORE_CELL_METADATA
                }

            if ref_metadata != test_metadata:
                if raise_on_first_difference:
                    try:
                        compare(test_metadata, ref_metadata)
                    except AssertionError as error:
                        raise NotebookDifference(
                            "Metadata differ on {} cell #{}: {}\nCell content:\n{}".format(
                                test_cell.cell_type, i, str(error), ref_cell.source
                            )
                        )
                else:
                    modified_cell_metadata.update(
                        set(test_metadata).difference(ref_metadata)
                    )
                    modified_cell_metadata.update(
                        set(ref_metadata).difference(test_metadata)
                    )
                    for key in set(ref_metadata).intersection(test_metadata):
                        if ref_metadata[key] != test_metadata[key]:
                            modified_cell_metadata.add(key)

        test_lines.extend(
            [
                line
                for line in test_cell.source.splitlines()
                if not _BLANK_LINE.match(line)
            ]
        )

        # 3. test cell content
        if ref_lines != test_lines:
            if raise_on_first_difference:
                try:
                    compare("\n".join(test_lines), "\n".join(ref_lines))
                except AssertionError as error:
                    raise NotebookDifference(
                        "Cell content differ on {} cell #{}: {}".format(
                            test_cell.cell_type, i, str(error)
                        )
                    )
            else:
                modified_cells.add(i)

        # 3. bis test entire cell content
        if not same_content(
            ref_cell.source, test_cell.source, allow_removed_final_blank_line
        ):
            if ref_cell.source != test_cell.source:
                if raise_on_first_difference:
                    diff = compare(test_cell.source, ref_cell.source, return_diff=True)
                    raise NotebookDifference(
                        "Cell content differ on {} cell #{}: {}".format(
                            test_cell.cell_type, i, diff
                        )
                    )
                modified_cells.add(i)

        if not compare_outputs:
            continue

        if ref_cell.cell_type != "code":
            continue

        ref_cell = filtered_cell(
            ref_cell,
            preserve_outputs=compare_outputs,
            cell_metadata_filter=cell_metadata_filter,
        )
        test_cell = filtered_cell(
            test_cell,
            preserve_outputs=compare_outputs,
            cell_metadata_filter=cell_metadata_filter,
        )

        try:
            compare(test_cell, ref_cell)
        except AssertionError as error:
            if raise_on_first_difference:
                raise NotebookDifference(
                    "Cell outputs differ on {} cell #{}: {}".format(
                        test_cell["cell_type"], i, str(error)
                    )
                )
            modified_cells.add(i)

    # More cells in the actual notebook?
    remaining_cell_count = 0
    while True:
        try:
            test_cell = next(test_cell_iter)
            if raise_on_first_difference:
                raise NotebookDifference(
                    "Additional {} cell: {}".format(
                        test_cell.cell_type, test_cell.source
                    )
                )
            remaining_cell_count += 1
        except StopIteration:
            break

    if remaining_cell_count and not raise_on_first_difference:
        modified_cells.update(
            range(
                len(expected_cells) + 1,
                len(expected_cells) + 1 + remaining_cell_count,
            )
        )

    return modified_cells, modified_cell_metadata


def test_round_trip_conversion(
    notebook, fmt, update, allow_expected_differences=True, stop_on_first_error=True
):
    """Test round trip conversion for a Jupyter notebook"""
    text = writes(notebook, fmt)
    round_trip = reads(text, fmt)

    if update:
        round_trip = combine_inputs_with_outputs(round_trip, notebook, fmt)

    compare_notebooks(
        round_trip,
        notebook,
        fmt,
        allow_expected_differences,
        raise_on_first_difference=stop_on_first_error,
    )



"""Find kernel specifications for a given language"""

import os
import sys

from .languages import same_language
from .reraise import reraise

try:
    # I prefer not to take a dependency on jupyter_client
    from jupyter_client.kernelspec import find_kernel_specs, get_kernel_spec
except ImportError as err:
    find_kernel_specs = reraise(err)
    get_kernel_spec = reraise(err)


def set_kernelspec_from_language(notebook):
    """Set the kernel specification based on the 'main_language' metadata"""
    language = notebook.metadata.get("jupytext", {}).get("main_language")
    if "kernelspec" not in notebook.metadata and language:
        try:
            kernelspec = kernelspec_from_language(language)
        except ValueError:
            return
        notebook.metadata["kernelspec"] = kernelspec
        notebook.metadata.get("jupytext", {}).pop("main_language")


def kernelspec_from_language(language):
    """Return the python kernel that matches the current env, or the first kernel that matches the given language"""
    if language == "python":
        # Return the kernel that matches the current Python executable
        for name in find_kernel_specs():
            kernel_specs = get_kernel_spec(name)
            cmd = kernel_specs.argv[0]
            if (
                kernel_specs.language == "python"
                and os.path.isfile(cmd)
                and os.path.samefile(cmd, sys.executable)
            ):
                return {
                    "name": name,
                    "language": language,
                    "display_name": kernel_specs.display_name,
                }
        raise ValueError(
            "No kernel found that matches the current python executable {}\n".format(
                sys.executable
            )
            + "Install one with 'python -m ipykernel install --name kernel_name [--user]'"
        )

    for name in find_kernel_specs():
        kernel_specs = get_kernel_spec(name)
        if same_language(kernel_specs.language, language):
            return {
                "name": name,
                "language": language,
                "display_name": kernel_specs.display_name,
            }

    raise ValueError("No kernel found for the language {}".format(language))



"""Raise the given error at evaluation time"""


def reraise(error):
    """Return a function that raises the given error when evaluated"""

    def local_function(*args, **kwargs):
        raise error

    return local_function



import pytest
from nbformat.v4.nbbase import new_notebook

import jupytext
from jupytext.compare import compare
from jupytext.formats import (
    JupytextFormatError,
    divine_format,
    get_format_implementation,
    guess_format,
    long_form_multiple_formats,
    read_format_from_metadata,
    rearrange_jupytext_metadata,
    short_form_multiple_formats,
    update_jupytext_formats_metadata,
    validate_one_format,
)

from .utils import list_notebooks, requires_myst, requires_pandoc


@pytest.mark.parametrize("nb_file", list_notebooks("python"))
def test_guess_format_light(nb_file):
    with open(nb_file) as stream:
        assert guess_format(stream.read(), ext=".py")[0] == "light"


@pytest.mark.parametrize("nb_file", list_notebooks("percent"))
def test_guess_format_percent(nb_file):
    with open(nb_file) as stream:
        assert guess_format(stream.read(), ext=".py")[0] == "percent"


def test_guess_format_simple_percent(
    nb="""# %%
print("hello world!")
""",
):
    assert guess_format(nb, ext=".py")[0] == "percent"


def test_guess_format_simple_percent_with_magic(
    nb="""# %%
# %time
print("hello world!")
""",
):
    assert guess_format(nb, ext=".py")[0] == "percent"


def test_guess_format_simple_hydrogen_with_magic(
    nb="""# %%
%time
print("hello world!")
""",
):
    assert guess_format(nb, ext=".py")[0] == "hydrogen"


@pytest.mark.parametrize("nb_file", list_notebooks("sphinx"))
def test_guess_format_sphinx(nb_file):
    with open(nb_file) as stream:
        assert guess_format(stream.read(), ext=".py")[0] == "sphinx"


def test_guess_format_hydrogen():
    text = """# %%
cat hello.txt
"""
    assert guess_format(text, ext=".py")[0] == "hydrogen"


def test_divine_format():
    assert divine_format('{"cells":[]}') == "ipynb"
    assert (
        divine_format(
            """def f(x):
    x + 1"""
        )
        == "py:light"
    )
    assert (
        divine_format(
            """# %%
def f(x):
    x + 1

# %%
def g(x):
    x + 2
"""
        )
        == "py:percent"
    )
    assert (
        divine_format(
            """This is a markdown file
with one code block

```
1 + 1
```
"""
        )
        == "md"
    )
    assert (
        divine_format(
            """;; ---
;; jupyter:
;;   jupytext:
;;     text_representation:
;;       extension: .ss
;;       format_name: percent
;; ---"""
        )
        == "ss:percent"
    )


def test_get_format_implementation():
    assert get_format_implementation(".py").format_name == "light"
    assert get_format_implementation(".py", "percent").format_name == "percent"
    with pytest.raises(JupytextFormatError):
        get_format_implementation(".py", "wrong_format")


def test_script_with_magics_not_percent(
    script="""# %%time
1 + 2""",
):
    assert guess_format(script, ".py")[0] == "light"


def test_script_with_spyder_cell_is_percent(
    script="""#%%
1 + 2""",
):
    assert guess_format(script, ".py")[0] == "percent"


def test_script_with_percent_cell_and_magic_is_hydrogen(
    script="""#%%
%matplotlib inline
""",
):
    assert guess_format(script, ".py")[0] == "hydrogen"


def test_script_with_percent_cell_and_kernelspec(
    script="""# ---
# jupyter:
#   kernelspec:
#     display_name: Python3
#     language: python
#     name: python3
# ---

# %%
a = 1
""",
):
    assert guess_format(script, ".py")[0] == "percent"


def test_script_with_spyder_cell_with_name_is_percent(
    script="""#%% cell name
1 + 2""",
):
    assert guess_format(script, ".py")[0] == "percent"


def test_read_format_from_metadata(
    script="""---
jupyter:
  jupytext:
    formats: ipynb,pct.py:percent,lgt.py:light,spx.py:sphinx,md,Rmd
    text_representation:
      extension: .pct.py
      format_name: percent
      format_version: '1.1'
      jupytext_version: 0.8.0
---""",
):
    assert read_format_from_metadata(script, ".Rmd") is None


def test_update_jupytext_formats_metadata():
    nb = new_notebook(metadata={"jupytext": {"formats": "py"}})
    update_jupytext_formats_metadata(nb.metadata, "py:light")
    assert nb.metadata["jupytext"]["formats"] == "py:light"

    nb = new_notebook(metadata={"jupytext": {"formats": "ipynb,py"}})
    update_jupytext_formats_metadata(nb.metadata, "py:light")
    assert nb.metadata["jupytext"]["formats"] == "ipynb,py:light"


def test_decompress_formats():
    assert long_form_multiple_formats("ipynb") == [{"extension": ".ipynb"}]
    assert long_form_multiple_formats("ipynb,md") == [
        {"extension": ".ipynb"},
        {"extension": ".md"},
    ]
    assert long_form_multiple_formats("ipynb,py:light") == [
        {"extension": ".ipynb"},
        {"extension": ".py", "format_name": "light"},
    ]
    assert long_form_multiple_formats(["ipynb", ".py:light"]) == [
        {"extension": ".ipynb"},
        {"extension": ".py", "format_name": "light"},
    ]
    assert long_form_multiple_formats(".pct.py:percent") == [
        {"extension": ".py", "suffix": ".pct", "format_name": "percent"}
    ]


def test_compress_formats():
    assert short_form_multiple_formats([{"extension": ".ipynb"}]) == "ipynb"
    assert short_form_multiple_formats("ipynb") == "ipynb"
    assert (
        short_form_multiple_formats([{"extension": ".ipynb"}, {"extension": ".md"}])
        == "ipynb,md"
    )
    assert (
        short_form_multiple_formats(
            [{"extension": ".ipynb"}, {"extension": ".py", "format_name": "light"}]
        )
        == "ipynb,py:light"
    )
    assert (
        short_form_multiple_formats(
            [
                {"extension": ".ipynb"},
                {"extension": ".py", "format_name": "light"},
                {"extension": ".md", "comment_magics": True},
            ]
        )
        == "ipynb,py:light,md"
    )
    assert (
        short_form_multiple_formats(
            [{"extension": ".py", "suffix": ".pct", "format_name": "percent"}]
        )
        == ".pct.py:percent"
    )


def test_rearrange_jupytext_metadata():
    metadata = {"nbrmd_formats": "ipynb,py"}
    rearrange_jupytext_metadata(metadata)
    compare(metadata, {"jupytext": {"formats": "ipynb,py"}})

    metadata = {"jupytext_formats": "ipynb,py"}
    rearrange_jupytext_metadata(metadata)
    compare(metadata, {"jupytext": {"formats": "ipynb,py"}})

    metadata = {"executable": "#!/bin/bash"}
    rearrange_jupytext_metadata(metadata)
    compare(metadata, {"jupytext": {"executable": "#!/bin/bash"}})


def test_rearrange_jupytext_metadata_metadata_filter():
    metadata = {
        "jupytext": {
            "metadata_filter": {
                "notebook": {"additional": ["one", "two"], "excluded": "all"},
                "cells": {"additional": "all", "excluded": ["three", "four"]},
            }
        }
    }
    rearrange_jupytext_metadata(metadata)
    compare(
        metadata,
        {
            "jupytext": {
                "notebook_metadata_filter": "one,two,-all",
                "cell_metadata_filter": "all,-three,-four",
            }
        },
    )


def test_rearrange_jupytext_metadata_add_dot_in_suffix():
    metadata = {
        "jupytext": {
            "text_representation": {"jupytext_version": "0.8.6"},
            "formats": "ipynb,pct.py,lgt.py",
        }
    }
    rearrange_jupytext_metadata(metadata)
    compare(
        metadata,
        {
            "jupytext": {
                "text_representation": {"jupytext_version": "0.8.6"},
                "formats": "ipynb,.pct.py,.lgt.py",
            }
        },
    )


def test_fix_139():
    text = """# ---
# jupyter:
#   jupytext:
#     metadata_filter:
#       cells:
#         additional:
#           - "lines_to_next_cell"
#         excluded:
#           - "all"
# ---

# + {"lines_to_next_cell": 2}
1 + 1
# -


1 + 1
"""

    nb = jupytext.reads(text, "py:light")
    text2 = jupytext.writes(nb, "py:light")
    assert "cell_metadata_filter: -all" in text2
    assert "lines_to_next_cell" not in text2


def test_validate_one_format():
    with pytest.raises(JupytextFormatError):
        validate_one_format("py:percent")

    with pytest.raises(JupytextFormatError):
        validate_one_format({"extension": "py", "format_name": "invalid"})

    with pytest.raises(JupytextFormatError):
        validate_one_format({})

    with pytest.raises(JupytextFormatError):
        validate_one_format({"extension": ".py", "unknown_option": True})

    with pytest.raises(JupytextFormatError):
        validate_one_format({"extension": ".py", "comment_magics": "TRUE"})


def test_set_auto_ext():
    with pytest.raises(ValueError):
        long_form_multiple_formats("ipynb,auto:percent", {})


@requires_pandoc
def test_pandoc_format_is_preserved():
    formats_org = "ipynb,md,.pandoc.md:pandoc,py:light"
    long = long_form_multiple_formats(formats_org)
    formats_new = short_form_multiple_formats(long)

    compare(formats_new, formats_org)


@requires_myst
def test_write_as_myst(tmpdir):
    """Inspired by https://github.com/mwouts/jupytext/issues/462"""
    nb = new_notebook()
    tmp_md = str(tmpdir.join("notebook.md"))

    jupytext.write(nb, tmp_md, fmt="myst")

    with open(tmp_md) as fp:
        md = fp.read()

    assert "myst" in md


def test_write_raises_when_fmt_does_not_exists(tmpdir):
    """Inspired by https://github.com/mwouts/jupytext/issues/462"""
    nb = new_notebook()
    tmp_md = str(tmpdir.join("notebook.md"))

    with pytest.raises(JupytextFormatError):
        jupytext.write(nb, tmp_md, fmt="unknown_format")



# ---
# jupyter:
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# # A quick insight at world population
#
# ## Collecting population data
#
# In the below we retrieve population data from the
# [World Bank](http://www.worldbank.org/)
# using the [wbdata](https://github.com/OliverSherouse/wbdata) python package

# +
import pandas as pd
import wbdata as wb

pd.options.display.max_rows = 6
pd.options.display.max_columns = 20
# -

# Corresponding indicator is found using search method - or, directly,
# the World Bank site.

wb.search_indicators('Population, total')  # SP.POP.TOTL
# wb.search_indicators('area')
# => https://data.worldbank.org/indicator is easier to use

# Now we download the population data

indicators = {'SP.POP.TOTL': 'Population, total',
              'AG.SRF.TOTL.K2': 'Surface area (sq. km)',
              'AG.LND.TOTL.K2': 'Land area (sq. km)',
              'AG.LND.ARBL.ZS': 'Arable land (% of land area)'}
data = wb.get_dataframe(indicators, convert_date=True).sort_index()
data

# World is one of the countries

data.loc['World']

# Can we classify over continents?

data.loc[(slice(None), '2017-01-01'), :]['Population, total'].dropna(
).sort_values().tail(60).index.get_level_values('country')

# Extract zones manually (in order of increasing population)

zones = ['North America', 'Middle East & North Africa',
         'Latin America & Caribbean', 'Europe & Central Asia',
         'Sub-Saharan Africa', 'South Asia',
         'East Asia & Pacific'][::-1]

# And extract population information (and check total is right)

population = data.loc[zones]['Population, total'].swaplevel().unstack()
population = population[zones]
assert all(data.loc['World']['Population, total'] == population.sum(axis=1))

# ## Stacked area plot with matplotlib

import matplotlib.pyplot as plt

plt.clf()
plt.figure(figsize=(10, 5), dpi=100)
plt.stackplot(population.index, population.values.T / 1e9)
plt.legend(population.columns, loc='upper left')
plt.ylabel('Population count (B)')
plt.show()

# ## Stacked bar plot with plotly

# Stacked area plots (with cumulated values computed depending on
# selected legends) are
# [on their way](https://github.com/plotly/plotly.js/pull/2960) at Plotly. For
# now we just do a stacked bar plot.

# +
import plotly.offline as offline
import plotly.graph_objs as go

offline.init_notebook_mode()
# -

bars = [go.Bar(x=population.index, y=population[zone], name=zone)
        for zone in zones]
fig = go.Figure(data=bars,
                layout=go.Layout(title='World population',
                                 barmode='stack'))
offline.iplot(fig)



from nbformat.v4.nbbase import new_code_cell, new_notebook

import jupytext
from jupytext.compare import compare, compare_notebooks


def test_hide_code_tag(
    no_jupytext_version_number,
    nb=new_notebook(
        metadata={"jupytext": {"main_language": "matlab"}},
        cells=[new_code_cell("1 + 1", metadata={"tags": ["hide_code"]})],
    ),
    text="""% + tags=["hide_code"]
1 + 1
""",
):
    text2 = jupytext.writes(nb, "m")
    compare(text2, text)
    nb2 = jupytext.reads(text, "m")
    compare_notebooks(nb2, nb)


def test_hide_code_tag_percent_format(
    no_jupytext_version_number,
    nb=new_notebook(
        metadata={"jupytext": {"main_language": "matlab"}},
        cells=[new_code_cell("1 + 1", metadata={"tags": ["hide_code"]})],
    ),
    text="""% %% tags=["hide_code"]
1 + 1
""",
):
    text2 = jupytext.writes(nb, "m:percent")
    compare(text2, text)
    nb2 = jupytext.reads(text, "m:percent")
    compare_notebooks(nb2, nb)



from nbformat.v4.nbbase import new_code_cell, new_markdown_cell, new_notebook

import jupytext
from jupytext.compare import compare, compare_notebooks


def test_read_magics(text="// :vars\n"):
    nb = jupytext.reads(text, "rs")
    compare_notebooks(nb, new_notebook(cells=[new_code_cell(":vars")]))
    compare(jupytext.writes(nb, "rs"), text)


def test_read_simple_file(
    text="""println!("Hello world");
eprintln!("Hello error");
format!("Hello {}", "world")

// A Function
pub fn fib(x: i32) -> i32 {
    if x <= 2 {0} else {fib(x - 2) + fib(x - 1)}
}

// This is a
// Markdown cell

// This is a magic instruction
// :vars

// This is a rust identifier
::std::mem::drop
""",
):
    nb = jupytext.reads(text, "rs")
    compare_notebooks(
        nb,
        new_notebook(
            cells=[
                new_code_cell(
                    """println!("Hello world");
eprintln!("Hello error");
format!("Hello {}", "world")"""
                ),
                new_code_cell(
                    """// A Function
pub fn fib(x: i32) -> i32 {
    if x <= 2 {0} else {fib(x - 2) + fib(x - 1)}
}"""
                ),
                new_markdown_cell("This is a\nMarkdown cell"),
                new_code_cell(
                    """// This is a magic instruction
:vars"""
                ),
                new_code_cell(
                    """// This is a rust identifier
::std::mem::drop"""
                ),
            ]
        ),
    )
    compare(jupytext.writes(nb, "rs"), text)


def test_read_write_script_with_metadata_241(
    no_jupytext_version_number,
    rsnb="""#!/usr/bin/env scriptisto
// ---
// jupyter:
//   jupytext:
//     text_representation:
//       extension: .rs
//       format_name: light
//   kernelspec:
//     display_name: Rust
//     language: rust
//     name: rust
// ---

let mut a: i32 = 2;
a += 1;
""",
):
    nb = jupytext.reads(rsnb, "rs")
    assert "executable" in nb.metadata["jupytext"]
    rsnb2 = jupytext.writes(nb, "rs")

    compare(rsnb, rsnb2)



import unittest.mock as mock
from pathlib import Path

import pytest
from git import Repo
from nbformat.v4 import nbbase
from nbformat.v4.nbbase import (
    new_code_cell,
    new_markdown_cell,
    new_notebook,
    new_output,
)

import jupytext
from jupytext.cli import system

# Pytest's tmpdir is in /tmp (at least for me), so this helps avoiding interferences between
# global configuration on HOME and the test collection
jupytext.config.JUPYTEXT_CEILING_DIRECTORIES = ["/tmp/"]


@pytest.fixture
def no_jupytext_version_number():
    with mock.patch("jupytext.header.INSERT_AND_CHECK_VERSION_NUMBER", False):
        yield


@pytest.fixture
def tmp_repo(tmpdir):
    repo = Repo.init(str(tmpdir))
    return repo


@pytest.fixture
def cwd_tmpdir(tmpdir):
    # Run the whole test from inside tmpdir
    with tmpdir.as_cwd():
        yield tmpdir


@pytest.fixture
def jupytext_repo_root():
    """The local path of this repo, to use in .pre-commit-config.yaml in tests"""
    return str(Path(__file__).parent.parent.resolve())


@pytest.fixture
def jupytext_repo_rev(jupytext_repo_root):
    """The local revision of this repo, to use in .pre-commit-config.yaml in tests"""
    return system("git", "rev-parse", "HEAD", cwd=jupytext_repo_root).strip()


@pytest.fixture()
def python_notebook():
    return new_notebook(
        cells=[new_markdown_cell("A short notebook")],
        # We need a Python kernel here otherwise Jupytext is going to add
        # the information that this is a Python notebook when we sync the
        # .py and .ipynb files
        metadata={
            "kernelspec": {
                "display_name": "Python 3",
                "language": "python",
                "name": "python_kernel",
            }
        },
    )


@pytest.fixture()
def notebook_with_outputs():
    return new_notebook(
        cells=[
            new_code_cell(
                "1+1",
                execution_count=1,
                outputs=[
                    new_output(
                        data={"text/plain": ["2"]},
                        execution_count=1,
                        output_type="execute_result",
                    )
                ],
            )
        ],
        metadata={
            "kernelspec": {
                "display_name": "Python 3",
                "language": "python",
                "name": "python_kernel",
            }
        },
    )


"""To make sure that cell ids are distinct we use a global counter.
    This solves https://github.com/mwouts/jupytext/issues/747"""
global_cell_count = 0


def generate_corpus_id():
    global global_cell_count
    global_cell_count += 1
    return f"cell-{global_cell_count}"


nbbase.random_cell_id = generate_corpus_id



# ---
# jupyter:
#   kernel_info:
#     name: python3
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# %% outputHidden=false inputHidden=false tags=["parameters"]
param = 4

# %% outputHidden=false inputHidden=false
import pandas as pd

# %% outputHidden=false inputHidden=false
df = pd.DataFrame({'A': [1, 2], 'B': [3 + param, 4]},
                  index=pd.Index(['x0', 'x1'], name='x'))
df

# %% outputHidden=false inputHidden=false
%matplotlib inline
df.plot(kind='bar')



import pytest

import jupytext

from .utils import list_notebooks


@pytest.mark.parametrize("r_file", list_notebooks("R_spin"))
def test_jupytext_same_as_knitr_spin(r_file, tmpdir):
    nb = jupytext.read(r_file)
    rmd_jupytext = jupytext.writes(nb, "Rmd")

    # Rmd file generated with spin(hair='R/spin.R', knit=FALSE)
    rmd_file = r_file.replace("R_spin", "Rmd").replace(".R", ".Rmd")

    with open(rmd_file) as fp:
        rmd_spin = fp.read()

    assert rmd_spin == rmd_jupytext



import nbformat
from nbformat.v4.nbbase import new_markdown_cell, new_notebook

from jupytext.cli import jupytext as jupytext_cli
from jupytext.compare import compare, compare_notebooks


def test_metadata_filters_from_config(tmpdir):
    cfg_file = tmpdir.join("jupytext.toml")
    nb_file = tmpdir.join("notebook.ipynb")
    md_file = tmpdir.join("notebook.md")

    cfg_file.write(
        """notebook_metadata_filter = "-all"
cell_metadata_filter = "-all"
"""
    )
    nb = new_notebook(
        cells=[new_markdown_cell("A markdown cell")],
        metadata={
            "kernelspec": {
                "display_name": "Python [conda env:.conda-week1]",
                "language": "python",
                "name": "conda-env-.conda-week1-py",
            },
            "language_info": {
                "codemirror_mode": {"name": "ipython", "version": 3},
                "file_extension": ".py",
                "mimetype": "text/x-python",
                "name": "python",
                "nbconvert_exporter": "python",
                "pygments_lexer": "ipython3",
                "version": "3.8.3",
            },
            "nbsphinx": {"execute": "never"},
        },
    )
    nbformat.write(nb, str(nb_file))

    jupytext_cli([str(nb_file), "--to", "md"])
    md = md_file.read()

    compare(md, "A markdown cell\n")

    jupytext_cli([str(md_file), "--to", "notebook", "--update"])
    nb2 = nbformat.read(str(nb_file), as_version=4)
    compare_notebooks(nb2, nb)



import shutil

import pytest
from git.exc import HookExecutionError
from nbformat.v4.nbbase import new_markdown_cell
from pre_commit.main import main as pre_commit

from jupytext import read, write
from jupytext.cli import jupytext

from .utils import (
    skip_pre_commit_tests_on_windows,
    skip_pre_commit_tests_when_jupytext_folder_is_not_a_git_repo,
)


@skip_pre_commit_tests_on_windows
@skip_pre_commit_tests_when_jupytext_folder_is_not_a_git_repo
def test_pre_commit_hook_sync(
    tmpdir,
    cwd_tmpdir,
    tmp_repo,
    jupytext_repo_root,
    jupytext_repo_rev,
    python_notebook,
):
    pre_commit_config_yaml = f"""
repos:
- repo: {jupytext_repo_root}
  rev: {jupytext_repo_rev}
  hooks:
  - id: jupytext
    args: [--sync]
"""
    tmpdir.join(".pre-commit-config.yaml").write(pre_commit_config_yaml)

    tmp_repo.git.add(".pre-commit-config.yaml")
    pre_commit(["install", "--install-hooks", "-f"])

    # write a test notebook and sync it to py:percent
    nb = python_notebook
    write(nb, "test.ipynb")

    jupytext(["--set-formats", "ipynb,py:percent", "test.ipynb"])

    # try to commit it, should fail because the py version hasn't been added
    tmp_repo.git.add("test.ipynb")
    with pytest.raises(
        HookExecutionError,
        match="git add test.py",
    ):
        tmp_repo.index.commit("failing")

    # add the py file, now the commit will succeed
    tmp_repo.git.add("test.py")
    tmp_repo.index.commit("passing")
    assert "test.ipynb" in tmp_repo.tree()
    assert "test.py" in tmp_repo.tree()

    # modify the ipynb file
    nb = read("test.ipynb")
    nb.cells.append(new_markdown_cell("A new cell"))
    write(nb, "test.ipynb")

    tmp_repo.git.add("test.ipynb")

    # We try to commit one more time, this updates the py file
    with pytest.raises(
        HookExecutionError,
        match="files were modified by this hook",
    ):
        tmp_repo.index.commit("failing")

    # the text file has been updated
    assert "A new cell" in tmpdir.join("test.py").read()

    # trying to commit should fail again because we forgot to add the .py version
    with pytest.raises(HookExecutionError, match="git add test.py"):
        tmp_repo.index.commit("still failing")

    nb = read("test.ipynb")
    assert len(nb.cells) == 2

    # add the text file, now the commit will succeed
    tmp_repo.git.add("test.py")
    tmp_repo.index.commit("passing")

    # modify the .py file
    nb.cells.append(new_markdown_cell("A third cell"))
    write(nb, "test.py", fmt="py:percent")
    tmp_repo.git.add("test.py")

    # the pre-commit hook will update the .ipynb file
    with pytest.raises(HookExecutionError, match="git add test.ipynb"):
        tmp_repo.index.commit("failing")

    tmp_repo.git.add("test.ipynb")
    tmp_repo.index.commit("passing")

    nb = read("test.ipynb")
    assert len(nb.cells) == 3

    # finally we move the paired notebook to a subfolder
    tmpdir.mkdir("subfolder")
    shutil.move("test.py", "subfolder")
    shutil.move("test.ipynb", "subfolder")

    # adding both files works on the first commit as notebooks are in sync
    tmp_repo.git.add("subfolder/test.ipynb")
    tmp_repo.git.add("subfolder/test.py")
    tmp_repo.index.commit("passing")

    # and we don't get any error or message when we run 'pre-commit run --all'
    status = pre_commit(["run", "--all"])
    assert status == 0



"""Issue #712"""
from nbformat.v4.nbbase import new_code_cell, new_notebook

from jupytext import reads, writes
from jupytext.cell_to_text import three_backticks_or_more
from jupytext.compare import compare, compare_notebooks

from .utils import requires_myst


def test_three_backticks_or_more():
    assert three_backticks_or_more([""]) == "```"
    assert three_backticks_or_more(["``"]) == "```"
    assert three_backticks_or_more(["```python"]) == "````"
    assert three_backticks_or_more(["```"]) == "````"
    assert three_backticks_or_more(["`````python"]) == "``````"
    assert three_backticks_or_more(["`````"]) == "``````"


def test_triple_backticks_in_code_cell(
    no_jupytext_version_number,
    nb=new_notebook(
        metadata={"main_language": "python"},
        cells=[
            new_code_cell(
                '''a = """
```
foo
```
"""'''
            )
        ],
    ),
    text='''---
jupyter:
  jupytext:
    main_language: python
---

````python
a = """
```
foo
```
"""
````
''',
):
    actual_text = writes(nb, fmt="md")
    compare(actual_text, text)

    actual_nb = reads(text, fmt="md")
    compare_notebooks(actual_nb, nb)


@requires_myst
def test_triple_backticks_in_code_cell_myst(
    no_jupytext_version_number,
    nb=new_notebook(
        metadata={"main_language": "python"},
        cells=[
            new_code_cell(
                '''a = """
```
foo
```
"""'''
            )
        ],
    ),
    text='''---
jupytext:
  main_language: python
---

````{code-cell}
a = """
```
foo
```
"""
````
''',
):
    actual_text = writes(nb, fmt="md:myst")
    compare(actual_text, text)

    actual_nb = reads(text, fmt="md:myst")
    compare_notebooks(actual_nb, nb)


def test_alternate_tree_four_five_backticks(
    no_jupytext_version_number,
    nb=new_notebook(
        metadata={"main_language": "python"},
        cells=[
            new_code_cell('a = """\n```\n"""'),
            new_code_cell("b = 2"),
            new_code_cell('c = """\n````\n"""'),
        ],
    ),
    text='''---
jupyter:
  jupytext:
    main_language: python
---

````python
a = """
```
"""
````

```python
b = 2
```

`````python
c = """
````
"""
`````
''',
):
    actual_text = writes(nb, fmt="md")
    compare(actual_text, text)

    actual_nb = reads(text, fmt="md")
    compare_notebooks(actual_nb, nb)



import os

import pytest

from jupytext.config import (
    find_jupytext_configuration_file,
    load_jupytext_configuration_file,
)


def test_find_jupytext_configuration_file(tmpdir):
    nested = tmpdir.mkdir("nested")

    # Start with no configuration
    assert find_jupytext_configuration_file(str(nested)) is None

    # Configuration file in the parent directory
    root_config = tmpdir.join("jupytext.yml")
    root_config.write("\n")
    assert os.path.samefile(
        find_jupytext_configuration_file(str(tmpdir)), str(root_config)
    )
    assert os.path.samefile(
        find_jupytext_configuration_file(str(nested)), str(root_config)
    )

    # Local pyproject file
    pyproject_config = nested.join("pyproject.toml")
    pyproject_config.write("[tool.jupytext]\n")
    assert os.path.samefile(
        find_jupytext_configuration_file(str(tmpdir)), str(root_config)
    )
    assert os.path.samefile(
        find_jupytext_configuration_file(str(nested)), str(pyproject_config)
    )

    # Local configuration file
    local_config = nested.join(".jupytext")
    local_config.write("\n")
    assert os.path.samefile(
        find_jupytext_configuration_file(str(tmpdir)), str(root_config)
    )
    assert os.path.samefile(
        find_jupytext_configuration_file(str(nested)), str(local_config)
    )


def test_jupytext_py_is_not_a_configuration_file(tmpdir):
    jupytext_py = tmpdir.join("jupytext.py")
    jupytext_py.write("# Not a config file!")

    assert find_jupytext_configuration_file(str(tmpdir)) is None

    dot_jupytext_py = tmpdir.join(".jupytext.py")
    dot_jupytext_py.write("# This is a config file!")

    assert find_jupytext_configuration_file(str(tmpdir)) == str(dot_jupytext_py)


@pytest.mark.parametrize(
    "config_file",
    [
        "pyproject.toml",
        "jupytext",
        "jupytext.toml",
        "jupytext.yml",
        "jupytext.json",
        "jupytext.py",
    ],
)
def test_load_jupytext_configuration_file(tmpdir, config_file):
    full_config_path = tmpdir.join(config_file)

    if config_file == "pyproject.toml":
        full_config_path.write(
            """[tool.jupytext]
formats = "ipynb,py:percent"
notebook_metadata_filter = "all"
cell_metadata_filter = "all"
"""
        )
    elif config_file.endswith(("jupytext", ".toml")):
        full_config_path.write(
            """formats = "ipynb,py:percent"
notebook_metadata_filter = "all"
cell_metadata_filter = "all"
"""
        )
    elif config_file.endswith(".yml"):
        full_config_path.write(
            """formats: ipynb,py:percent
notebook_metadata_filter: all
cell_metadata_filter: all
"""
        )
    elif config_file.endswith(".json"):
        full_config_path.write(
            """{"formats": "ipynb,py:percent",
"notebook_metadata_filter": "all",
"cell_metadata_filter": "all"
}
"""
        )
    elif config_file.endswith(".py"):
        full_config_path.write(
            """c.formats = "ipynb,py:percent"
c.notebook_metadata_filter = "all"
c.cell_metadata_filter = "all"
"""
        )

    config = load_jupytext_configuration_file(str(full_config_path))
    assert config.formats == "ipynb,py:percent"
    assert config.notebook_metadata_filter == "all"
    assert config.cell_metadata_filter == "all"


@pytest.mark.parametrize(
    "content_toml,formats_short_form",
    [
        (
            """# always pair ipynb notebooks to py:percent files
formats = "ipynb,py:percent"
""",
            "ipynb,py:percent",
        ),
        (
            """# always pair ipynb notebooks to py:percent files
formats = ['ipynb', 'py:percent']
""",
            "ipynb,py:percent",
        ),
        (
            """# always pair ipynb notebooks to py:percent files
formats = ["ipynb", "py:percent"]
""",
            "ipynb,py:percent",
        ),
        (
            """# always pair ipynb notebooks to py:percent files
formats = "ipynb,py:percent"
""",
            "ipynb,py:percent",
        ),
        (
            """# Pair notebooks in subfolders of 'notebooks' to scripts in subfolders of 'scripts'
[formats]
"notebooks/" = "ipynb"
"scripts/" = "py:percent"
""",
            "notebooks///ipynb,scripts///py:percent",
        ),
        (
            """# Pair notebooks in subfolders of 'notebooks' to scripts in subfolders of 'scripts'
[formats]
"notebooks" = "ipynb"
"scripts" = "py:percent"
""",
            "notebooks///ipynb,scripts///py:percent",
        ),
    ],
)
def test_jupytext_formats(tmpdir, content_toml, formats_short_form):
    jupytext_toml = tmpdir.join("jupytext.toml")
    jupytext_toml.write(content_toml)

    config = load_jupytext_configuration_file(str(jupytext_toml))
    assert config.formats == formats_short_form


def test_deprecated_formats_cause_warning(
    tmpdir, content_toml="default_jupytext_formats = 'ipynb,md'"
):
    jupytext_toml = tmpdir.join("jupytext.toml")
    jupytext_toml.write(content_toml)

    config = load_jupytext_configuration_file(str(jupytext_toml))
    with pytest.warns(FutureWarning, match="use 'formats'"):
        assert config.default_formats(str(tmpdir.join("test.md"))) == "ipynb,md"


@pytest.mark.parametrize(
    "option_name",
    ["notebook_metadata_filter", "cell_metadata_filter", "cell_markers"],
)
def test_deprecated_options_cause_warning(tmpdir, option_name):
    jupytext_toml = tmpdir.join("jupytext.toml")
    jupytext_toml.write(f"default_{option_name} = 'value'")
    config = load_jupytext_configuration_file(str(jupytext_toml))
    fmt = {}
    with pytest.warns(FutureWarning, match=f"use '{option_name}'"):
        config.set_default_format_options(fmt)
        assert fmt[option_name] == "value"



# -*- coding: utf-8 -*-

import pytest
from nbformat.v4.nbbase import (
    new_code_cell,
    new_markdown_cell,
    new_notebook,
    new_raw_cell,
)

import jupytext
from jupytext.compare import compare, compare_notebooks


def test_read_simple_file(
    pynb="""# ---
# title: Simple file
# ---

# Here we have some text
# And below we have some python code

def f(x):
    return x+1


def h(y):
    return y-1
""",
):
    nb = jupytext.reads(pynb, "py")
    assert len(nb.cells) == 4
    assert nb.cells[0].cell_type == "raw"
    assert nb.cells[0].source == "---\ntitle: Simple file\n---"
    assert nb.cells[1].cell_type == "markdown"
    assert (
        nb.cells[1].source == "Here we have some text\n"
        "And below we have some python code"
    )
    assert nb.cells[2].cell_type == "code"
    compare(
        nb.cells[2].source,
        """def f(x):
    return x+1""",
    )
    assert nb.cells[3].cell_type == "code"
    compare(
        nb.cells[3].source,
        """def h(y):
    return y-1""",
    )

    pynb2 = jupytext.writes(nb, "py")
    compare(pynb2, pynb)


def test_read_less_simple_file(
    pynb="""# ---
# title: Less simple file
# ---

# Here we have some text
# And below we have some python code

# This is a comment about function f
def f(x):
    return x+1


# And a comment on h
def h(y):
    return y-1
""",
):
    nb = jupytext.reads(pynb, "py")

    assert len(nb.cells) == 4
    assert nb.cells[0].cell_type == "raw"
    assert nb.cells[0].source == "---\ntitle: Less simple file\n---"
    assert nb.cells[1].cell_type == "markdown"
    assert (
        nb.cells[1].source == "Here we have some text\n"
        "And below we have some python code"
    )
    assert nb.cells[2].cell_type == "code"
    compare(
        nb.cells[2].source,
        "# This is a comment about function f\n" "def f(x):\n" "    return x+1",
    )
    assert nb.cells[3].cell_type == "code"
    compare(nb.cells[3].source, """# And a comment on h\ndef h(y):\n    return y-1""")

    pynb2 = jupytext.writes(nb, "py")
    compare(pynb2, pynb)


def test_indented_comment(
    text="""def f():
    return 1

    # f returns 1


def g():
    return 2


# h returns 3
def h():
    return 3
""",
    ref=new_notebook(
        cells=[
            new_code_cell(
                """def f():
    return 1

    # f returns 1"""
            ),
            new_code_cell("def g():\n    return 2"),
            new_code_cell("# h returns 3\ndef h():\n    return 3"),
        ]
    ),
):
    nb = jupytext.reads(text, "py")
    compare_notebooks(nb, ref)
    py = jupytext.writes(nb, "py")
    compare(py, text)


def test_non_pep8(
    text="""def f():
    return 1
def g():
    return 2

def h():
    return 3
""",
    ref=new_notebook(
        cells=[
            new_code_cell(
                "def f():\n    return 1\ndef g():\n    return 2",
                metadata={"lines_to_next_cell": 1},
            ),
            new_code_cell("def h():\n    return 3"),
        ]
    ),
):
    nb = jupytext.reads(text, "py")
    compare_notebooks(nb, ref)
    py = jupytext.writes(nb, "py")
    compare(py, text)


def test_read_non_pep8_file(
    pynb="""# ---
# title: Non-pep8 file
# ---

# This file is non-pep8 as the function below has
# two consecutive blank lines in its body

def f(x):


    return x+1
""",
):
    nb = jupytext.reads(pynb, "py")

    assert len(nb.cells) == 3
    assert nb.cells[0].cell_type == "raw"
    assert nb.cells[0].source == "---\ntitle: Non-pep8 file\n---"
    assert nb.cells[1].cell_type == "markdown"
    assert (
        nb.cells[1].source == "This file is non-pep8 as "
        "the function below has\n"
        "two consecutive blank lines "
        "in its body"
    )
    assert nb.cells[2].cell_type == "code"
    compare(nb.cells[2].source, "def f(x):\n\n\n" "    return x+1")

    pynb2 = jupytext.writes(nb, "py")
    compare(pynb2, pynb)


def test_read_cell_two_blank_lines(
    pynb="""# ---
# title: cell with two consecutive blank lines
# ---

# +
a = 1


a + 2
""",
):
    nb = jupytext.reads(pynb, "py")

    assert len(nb.cells) == 2
    assert nb.cells[0].cell_type == "raw"
    assert (
        nb.cells[0].source == "---\ntitle: cell with two "
        "consecutive blank lines\n---"
    )
    assert nb.cells[1].cell_type == "code"
    assert nb.cells[1].source == "a = 1\n\n\na + 2"

    pynb2 = jupytext.writes(nb, "py")
    compare(pynb2, pynb)


def test_read_cell_explicit_start(
    pynb="""
import pandas as pd
# +
def data():
    return pd.DataFrame({'A': [0, 1]})


data()
""",
):
    nb = jupytext.reads(pynb, "py")
    pynb2 = jupytext.writes(nb, "py")
    compare(pynb2, pynb)


def test_read_complex_cells(
    pynb="""import pandas as pd

# +
def data():
    return pd.DataFrame({'A': [0, 1]})


data()

# +
def data2():
    return pd.DataFrame({'B': [0, 1]})


data2()

# +
# Finally we have a cell with only comments
# This cell should remain a code cell and not get converted
# to markdown

# + {"endofcell": "--"}
# This cell has an enumeration in it that should not
# match the endofcell marker!
# - item 1
# - item 2
# -
# --
""",
):
    nb = jupytext.reads(pynb, "py")
    assert len(nb.cells) == 5
    assert nb.cells[0].cell_type == "code"
    assert nb.cells[1].cell_type == "code"
    assert nb.cells[2].cell_type == "code"
    assert nb.cells[3].cell_type == "code"
    assert nb.cells[4].cell_type == "code"
    assert (
        nb.cells[3].source
        == """# Finally we have a cell with only comments
# This cell should remain a code cell and not get converted
# to markdown"""
    )
    assert (
        nb.cells[4].source
        == """# This cell has an enumeration in it that should not
# match the endofcell marker!
# - item 1
# - item 2
# -"""
    )

    pynb2 = jupytext.writes(nb, "py")
    compare(pynb2, pynb)


def test_read_prev_function(
    pynb="""def test_read_cell_explicit_start_end(pynb='''
import pandas as pd
# +
def data():
    return pd.DataFrame({'A': [0, 1]})


data()
'''):
    nb = jupytext.reads(pynb, 'py')
    pynb2 = jupytext.writes(nb, 'py')
    compare(pynb2, pynb)
""",
):
    nb = jupytext.reads(pynb, "py")
    pynb2 = jupytext.writes(nb, "py")
    compare(pynb2, pynb)


def test_read_cell_with_one_blank_line_end(
    pynb="""import pandas

""",
):
    nb = jupytext.reads(pynb, "py")
    assert len(nb.cells) == 1
    pynb2 = jupytext.writes(nb, "py")
    compare(pynb2, pynb)


def test_read_code_cell_fully_commented(
    pynb="""# +
# This is a code cell that
# only contains comments
""",
):
    nb = jupytext.reads(pynb, "py")
    assert len(nb.cells) == 1
    assert nb.cells[0].cell_type == "code"
    assert (
        nb.cells[0].source
        == """# This is a code cell that
# only contains comments"""
    )
    pynb2 = jupytext.writes(nb, "py")
    compare(pynb2, pynb)


def test_file_with_two_blank_line_end(
    pynb="""import pandas


""",
):
    nb = jupytext.reads(pynb, "py")
    pynb2 = jupytext.writes(nb, "py")
    compare(pynb2, pynb)


def test_one_blank_lines_after_endofcell(
    pynb="""# +
# This is a code cell with explicit end of cell
1 + 1

2 + 2
# -

# This cell is a cell with implicit start
1 + 1
""",
):
    nb = jupytext.reads(pynb, "py")
    assert len(nb.cells) == 2
    assert nb.cells[0].cell_type == "code"
    assert (
        nb.cells[0].source
        == """# This is a code cell with explicit end of cell
1 + 1

2 + 2"""
    )
    assert nb.cells[1].cell_type == "code"
    assert (
        nb.cells[1].source
        == """# This cell is a cell with implicit start
1 + 1"""
    )
    pynb2 = jupytext.writes(nb, "py")
    compare(pynb2, pynb)


def test_two_cells_with_explicit_start(
    pynb="""# +
# Cell one
1 + 1

1 + 1

# +
# Cell two
2 + 2

2 + 2
""",
):
    nb = jupytext.reads(pynb, "py")
    assert len(nb.cells) == 2
    assert nb.cells[0].cell_type == "code"
    assert (
        nb.cells[0].source
        == """# Cell one
1 + 1

1 + 1"""
    )
    assert nb.cells[1].cell_type == "code"
    assert (
        nb.cells[1].source
        == """# Cell two
2 + 2

2 + 2"""
    )
    pynb2 = jupytext.writes(nb, "py")
    compare(pynb2, pynb)


def test_escape_start_pattern(
    pynb="""# The code start pattern '# +' can
# appear in code and markdown cells.

# In markdown cells it is escaped like here:
# # + {"sample_metadata": "value"}

# In code cells like this one, it is also escaped
# # + {"sample_metadata": "value"}
1 + 1
""",
):
    nb = jupytext.reads(pynb, "py")
    assert len(nb.cells) == 3
    assert nb.cells[0].cell_type == "markdown"
    assert nb.cells[1].cell_type == "markdown"
    assert nb.cells[2].cell_type == "code"
    assert (
        nb.cells[1].source
        == """In markdown cells it is escaped like here:
# + {"sample_metadata": "value"}"""
    )
    assert (
        nb.cells[2].source
        == """# In code cells like this one, it is also escaped
# + {"sample_metadata": "value"}
1 + 1"""
    )
    pynb2 = jupytext.writes(nb, "py")
    compare(pynb2, pynb)


def test_dictionary_with_blank_lines_not_broken(
    pynb="""# This is a markdown cell, and below
# we have a long dictionary with blank lines
# inside it

dictionary = {
    'a': 'A',
    'b': 'B',

    # and the end
    'z': 'Z'}
""",
):
    nb = jupytext.reads(pynb, "py")
    assert len(nb.cells) == 2
    assert nb.cells[0].cell_type == "markdown"
    assert nb.cells[1].cell_type == "code"
    assert (
        nb.cells[0].source
        == """This is a markdown cell, and below
we have a long dictionary with blank lines
inside it"""
    )
    assert (
        nb.cells[1].source
        == """dictionary = {
    'a': 'A',
    'b': 'B',

    # and the end
    'z': 'Z'}"""
    )
    pynb2 = jupytext.writes(nb, "py")
    compare(pynb2, pynb)


def test_isolated_cell_with_magic(
    pynb="""# ---
# title: cell with isolated jupyter magic
# ---

# A magic command included in a markdown
# paragraph is code
#
# %matplotlib inline

# a code block may start with
# a magic command, like this one:

# %matplotlib inline

# or that one

# %matplotlib inline
1 + 1
""",
):
    nb = jupytext.reads(pynb, "py")

    assert len(nb.cells) == 6
    assert nb.cells[0].cell_type == "raw"
    assert nb.cells[0].source == "---\ntitle: cell with isolated jupyter " "magic\n---"
    assert nb.cells[1].cell_type == "code"
    assert nb.cells[2].cell_type == "markdown"
    assert nb.cells[3].cell_type == "code"
    assert nb.cells[3].source == "%matplotlib inline"
    assert nb.cells[4].cell_type == "markdown"
    assert nb.cells[5].cell_type == "code"
    assert nb.cells[5].source == "%matplotlib inline\n1 + 1"

    pynb2 = jupytext.writes(nb, "py")
    compare(pynb2, pynb)


def test_ipython_help_are_commented_297(
    text="""# This is a markdown cell
# that ends with a question: float?

# The next cell is also a markdown cell,
# because it has no code marker:

# float?

# +
# float?

# +
# float??

# +
# Finally a question in a code
# # cell?
""",
    nb=new_notebook(
        cells=[
            new_markdown_cell(
                "This is a markdown cell\nthat ends with a question: float?"
            ),
            new_markdown_cell(
                "The next cell is also a markdown cell,\nbecause it has no code marker:"
            ),
            new_markdown_cell("float?"),
            new_code_cell("float?"),
            new_code_cell("float??"),
            new_code_cell("# Finally a question in a code\n# cell?"),
        ]
    ),
):
    nb2 = jupytext.reads(text, "py")
    compare_notebooks(nb2, nb)

    text2 = jupytext.writes(nb2, "py")
    compare(text2, text)


def test_questions_in_unmarked_cells_are_not_uncommented_297(
    text="""# This cell has no explicit marker
# question?
1 + 2
""",
    nb=new_notebook(
        cells=[
            new_code_cell(
                "# This cell has no explicit marker\n# question?\n1 + 2",
                metadata={"comment_questions": False},
            )
        ]
    ),
):
    nb2 = jupytext.reads(text, "py")
    compare_notebooks(nb2, nb)

    text2 = jupytext.writes(nb2, "py")
    compare(text2, text)


def test_read_multiline_comment(
    pynb="""'''This is a multiline
comment with "quotes", 'single quotes'
# and comments
and line breaks


and it ends here'''


1 + 1
""",
):
    nb = jupytext.reads(pynb, "py")

    assert len(nb.cells) == 2
    assert nb.cells[0].cell_type == "code"
    assert (
        nb.cells[0].source
        == """'''This is a multiline
comment with "quotes", 'single quotes'
# and comments
and line breaks


and it ends here'''"""
    )
    assert nb.cells[1].cell_type == "code"
    assert nb.cells[1].source == "1 + 1"

    pynb2 = jupytext.writes(nb, "py")
    compare(pynb2, pynb)


def test_no_space_after_code(
    pynb=u"""# -*- coding: utf-8 -*-
# Markdown cell

def f(x):
    return x+1

# And a new cell, and non ascii contênt
""",
):
    nb = jupytext.reads(pynb, "py")

    assert len(nb.cells) == 3
    assert nb.cells[0].cell_type == "markdown"
    assert nb.cells[0].source == "Markdown cell"
    assert nb.cells[1].cell_type == "code"
    assert nb.cells[1].source == "def f(x):\n    return x+1"
    assert nb.cells[2].cell_type == "markdown"
    assert nb.cells[2].source == u"And a new cell, and non ascii contênt"

    pynb2 = jupytext.writes(nb, "py")
    compare(pynb2, pynb)


def test_read_write_script(
    pynb="""#!/usr/bin/env python
# coding=utf-8
print('Hello world')
""",
):
    nb = jupytext.reads(pynb, "py")
    pynb2 = jupytext.writes(nb, "py")
    compare(pynb2, pynb)


def test_read_write_script_with_metadata_241(
    no_jupytext_version_number,
    pynb="""#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ---
# jupyter:
#   jupytext:
#     text_representation:
#       extension: .py
#       format_name: light
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

a = 2

a + 1
""",
):
    nb = jupytext.reads(pynb, "py")
    assert "executable" in nb.metadata["jupytext"]
    assert "encoding" in nb.metadata["jupytext"]
    pynb2 = jupytext.writes(nb, "py")

    compare(pynb2, pynb)


def test_notebook_blank_lines(
    script="""# +
# This is a comment
# followed by two variables
a = 3

b = 4
# -

# New cell is a variable
c = 5


# +
# Now we have two functions
def f(x):
    return x + x


def g(x):
    return x + x + x


# -


# A commented block that is two lines away
# from previous cell

# A function again
def h(x):
    return x + 1


# variable
d = 6
""",
):
    notebook = jupytext.reads(script, "py")
    assert len(notebook.cells) >= 6
    for cell in notebook.cells:
        lines = cell.source.splitlines()
        if len(lines) != 1:
            assert lines[0], cell.source
            assert lines[-1], cell.source

    script2 = jupytext.writes(notebook, "py")

    compare(script2, script)


def test_notebook_two_blank_lines_before_next_cell(
    script="""# +
# This is cell with a function

def f(x):
    return 4


# +
# Another cell
c = 5


def g(x):
    return 6


# +
# Final cell

1 + 1
""",
):
    notebook = jupytext.reads(script, "py")
    assert len(notebook.cells) == 3
    for cell in notebook.cells:
        lines = cell.source.splitlines()
        if len(lines) != 1:
            assert lines[0]
            assert lines[-1]

    script2 = jupytext.writes(notebook, "py")

    compare(script2, script)


def test_notebook_one_blank_line_between_cells(
    script="""# +
1 + 1

2 + 2

# +
3 + 3

4 + 4

# +
5 + 5


def g(x):
    return 6


# +
7 + 7


def h(x):
    return 8


# +
def i(x):
    return 9


10 + 10


# +
def j(x):
    return 11


12 + 12
""",
):
    notebook = jupytext.reads(script, "py")
    for cell in notebook.cells:
        lines = cell.source.splitlines()
        assert lines[0]
        assert lines[-1]
        assert not cell.metadata, cell.source

    script2 = jupytext.writes(notebook, "py")

    compare(script2, script)


def test_notebook_with_magic_and_bash_cells(
    script="""# This is a test for issue #181

# %load_ext line_profiler

# !head -4 data/president_heights.csv
""",
):
    notebook = jupytext.reads(script, "py")
    for cell in notebook.cells:
        lines = cell.source.splitlines()
        assert lines[0]
        assert lines[-1]
        assert not cell.metadata, cell.source

    script2 = jupytext.writes(notebook, "py")

    compare(script2, script)


def test_notebook_no_line_to_next_cell(
    nb=new_notebook(
        cells=[
            new_markdown_cell("Markdown cell #1"),
            new_code_cell("%load_ext line_profiler"),
            new_markdown_cell("Markdown cell #2"),
            new_code_cell("%lprun -f ..."),
            new_markdown_cell("Markdown cell #3"),
            new_code_cell("# And a function!\n" "def f(x):\n" "    return 5"),
        ]
    )
):
    script = jupytext.writes(nb, "py")
    nb2 = jupytext.reads(script, "py")
    nb2.metadata.pop("jupytext")

    compare_notebooks(nb2, nb)


def test_notebook_one_blank_line_before_first_markdown_cell(
    script="""
# This is a markdown cell

1 + 1
""",
):
    notebook = jupytext.reads(script, "py")
    script2 = jupytext.writes(notebook, "py")
    compare(script2, script)

    assert len(notebook.cells) == 3
    for cell in notebook.cells:
        lines = cell.source.splitlines()
        if len(lines):
            assert lines[0]
            assert lines[-1]


def test_read_markdown_cell_with_triple_quote_307(
    script="""# This script test that commented triple quotes '''
# do not impede the correct identification of Markdown cells

# Here is Markdown cell number 2 '''
""",
):
    notebook = jupytext.reads(script, "py")
    assert len(notebook.cells) == 2
    assert notebook.cells[0].cell_type == "markdown"
    assert (
        notebook.cells[0].source
        == """This script test that commented triple quotes '''
do not impede the correct identification of Markdown cells"""
    )
    assert notebook.cells[1].cell_type == "markdown"
    assert notebook.cells[1].source == "Here is Markdown cell number 2 '''"

    script2 = jupytext.writes(notebook, "py")
    compare(script2, script)


def test_read_explicit_markdown_cell_with_triple_quote_307(
    script="""# {{{ [md] {"special": "metadata"}
# some text '''
# }}}

print('hello world')

# {{{ [md] {"special": "metadata"}
# more text '''
# }}}
""",
):
    notebook = jupytext.reads(script, "py")
    assert len(notebook.cells) == 3
    assert notebook.cells[0].cell_type == "markdown"
    assert notebook.cells[0].source == "some text '''"
    assert notebook.cells[1].cell_type == "code"
    assert notebook.cells[1].source == "print('hello world')"
    assert notebook.cells[2].cell_type == "markdown"
    assert notebook.cells[2].source == "more text '''"

    script2 = jupytext.writes(notebook, "py")
    compare(script2, script)


def test_round_trip_markdown_cell_with_magic():
    notebook = new_notebook(
        cells=[new_markdown_cell("IPython has magic commands like\n%quickref")],
        metadata={"jupytext": {"main_language": "python"}},
    )
    text = jupytext.writes(notebook, "py")
    notebook2 = jupytext.reads(text, "py")
    compare_notebooks(notebook2, notebook)


def test_round_trip_python_with_js_cell():
    notebook = new_notebook(
        cells=[
            new_code_cell(
                """import notebook.nbextensions
notebook.nbextensions.install_nbextension('index.js', user=True)"""
            ),
            new_code_cell(
                """%%javascript
Jupyter.utils.load_extensions('jupytext')"""
            ),
        ]
    )
    text = jupytext.writes(notebook, "py")
    notebook2 = jupytext.reads(text, "py")
    compare_notebooks(notebook2, notebook)


def test_round_trip_python_with_js_cell_no_cell_metadata():
    notebook = new_notebook(
        cells=[
            new_code_cell(
                """import notebook.nbextensions
notebook.nbextensions.install_nbextension('index.js', user=True)"""
            ),
            new_code_cell(
                """%%javascript
Jupyter.utils.load_extensions('jupytext')"""
            ),
        ],
        metadata={
            "jupytext": {
                "notebook_metadata_filter": "-all",
                "cell_metadata_filter": "-all",
            }
        },
    )
    text = jupytext.writes(notebook, "py")
    notebook2 = jupytext.reads(text, "py")
    compare_notebooks(notebook2, notebook)


def test_raw_with_metadata(
    no_jupytext_version_number,
    text="""# + key="value" active=""
# Raw cell
# # Commented line
""",
    notebook=new_notebook(
        cells=[new_raw_cell("Raw cell\n# Commented line", metadata={"key": "value"})]
    ),
):
    nb2 = jupytext.reads(text, "py")
    compare_notebooks(nb2, notebook)
    text2 = jupytext.writes(notebook, "py")
    compare(text2, text)


def test_raw_with_metadata_2(
    no_jupytext_version_number,
    text="""# + [raw] key="value"
# Raw cell
# # Commented line
""",
    notebook=new_notebook(
        cells=[new_raw_cell("Raw cell\n# Commented line", metadata={"key": "value"})]
    ),
):
    nb2 = jupytext.reads(text, "py")
    compare_notebooks(nb2, notebook)


def test_markdown_with_metadata(
    no_jupytext_version_number,
    text="""# + [markdown] key="value"
# Markdown cell
""",
    notebook=new_notebook(
        cells=[new_markdown_cell("Markdown cell", metadata={"key": "value"})]
    ),
):
    nb2 = jupytext.reads(text, "py")
    compare_notebooks(nb2, notebook)
    text2 = jupytext.writes(notebook, "py")
    compare(text2, text)


def test_multiline_comments_in_markdown_1():
    text = """# + [markdown]
'''
a
long
cell
'''
"""
    nb = jupytext.reads(text, "py")
    assert len(nb.cells) == 1
    assert nb.cells[0].cell_type == "markdown"
    assert nb.cells[0].source == "a\nlong\ncell"
    py = jupytext.writes(nb, "py")
    compare(py, text)


def test_multiline_comments_in_markdown_2():
    text = '''# + [markdown]
"""
a
long
cell
"""
'''
    nb = jupytext.reads(text, "py")
    assert len(nb.cells) == 1
    assert nb.cells[0].cell_type == "markdown"
    assert nb.cells[0].source == "a\nlong\ncell"
    py = jupytext.writes(nb, "py")
    compare(py, text)


def test_multiline_comments_in_raw_cell():
    text = '''# + active=""
"""
some
text
"""
'''
    nb = jupytext.reads(text, "py")
    assert len(nb.cells) == 1
    assert nb.cells[0].cell_type == "raw"
    assert nb.cells[0].source == "some\ntext"
    py = jupytext.writes(nb, "py")
    compare(py, text)


def test_multiline_comments_in_markdown_cell_no_line_return():
    text = '''# + [md]
"""a
long
cell"""
'''
    nb = jupytext.reads(text, "py")
    assert len(nb.cells) == 1
    assert nb.cells[0].cell_type == "markdown"
    assert nb.cells[0].source == "a\nlong\ncell"


def test_multiline_comments_in_markdown_cell_is_robust_to_additional_cell_marker():
    text = '''# + [md]
"""
some text, and a fake cell marker
# + [raw]
"""
'''
    nb = jupytext.reads(text, "py")
    assert len(nb.cells) == 1
    assert nb.cells[0].cell_type == "markdown"
    assert nb.cells[0].source == "some text, and a fake cell marker\n# + [raw]"
    py = jupytext.writes(nb, "py")
    compare(py, text)


def test_active_tag(
    text="""# + tags=["active-py"]
interpreter = 'python'

# + tags=["active-ipynb"]
# interpreter = 'ipython'
""",
    ref=new_notebook(
        cells=[
            new_raw_cell("interpreter = 'python'", metadata={"tags": ["active-py"]}),
            new_code_cell(
                "interpreter = 'ipython'", metadata={"tags": ["active-ipynb"]}
            ),
        ]
    ),
):
    nb = jupytext.reads(text, "py")
    compare_notebooks(nb, ref)
    py = jupytext.writes(nb, "py")
    compare(py, text)


def test_indented_bash_command(
    no_jupytext_version_number,
    nb=new_notebook(
        cells=[
            new_code_cell(
                """try:
    !echo jo
    pass
except:
    pass"""
            )
        ]
    ),
    text="""try:
    # !echo jo
    pass
except:
    pass
""",
):
    """Reproduces https://github.com/mwouts/jupytext/issues/437"""
    py = jupytext.writes(nb, "py")
    compare(py, text)
    nb2 = jupytext.reads(py, "py")
    compare_notebooks(nb2, nb)


def test_two_raw_cells_are_preserved(
    nb=new_notebook(cells=[new_raw_cell("---\nX\n---"), new_raw_cell("Y")])
):
    """Test the pattern described at https://github.com/mwouts/jupytext/issues/466"""
    py = jupytext.writes(nb, "py")
    nb2 = jupytext.reads(py, "py")
    compare_notebooks(nb2, nb)


def test_no_metadata_on_multiline_decorator(
    text="""import pytest


@pytest.mark.parametrize(
    "arg",
    [
        'a',
        'b',
        'c'
    ],
)
def test_arg(arg):
    assert isinstance(arg, str)
""",
):
    """Applying black on the code of jupytext 1.4.2 turns some pytest parameters into multi-lines ones, and
    causes a few failures in test_pep8.py:test_no_metadata_when_py_is_pep8"""
    nb = jupytext.reads(text, "py")
    assert len(nb.cells) == 2
    for cell in nb.cells:
        assert cell.cell_type == "code"
    assert nb.cells[0].source == "import pytest"
    assert nb.cells[0].metadata == {}


@pytest.mark.parametrize(
    "script,cell",
    [
        (
            """if True:
    # # !rm file 1
    # !rm file 2
""",
            """if True:
    # !rm file 1
    !rm file 2""",
        ),
        (
            """# +
if True:
    # help?
    # ?help
    # # ?help
    # # help?
""",
            """if True:
    help?
    ?help
    # ?help
    # help?""",
        ),
    ],
)
def test_indented_magic_commands(script, cell):

    nb = jupytext.reads(script, "py")
    assert len(nb.cells) == 1
    assert nb.cells[0].cell_type == "code"
    compare(nb.cells[0].source, cell)
    assert nb.cells[0].metadata == {}
    compare(jupytext.writes(nb, "py"), script)



# ---
# jupyter:
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# %% [markdown] slideshow={"slide_type": "slide"}
# A markdown cell

# %% slideshow={"slide_type": ""}
1+1

# %% [markdown] cell_style="center" slideshow={"slide_type": "fragment"}
# Markdown cell two



from copy import deepcopy

import pytest
from nbformat.v4.nbbase import new_code_cell, new_markdown_cell, new_notebook

import jupytext
from jupytext.combine import combine_inputs_with_outputs
from jupytext.compare import compare, compare_notebooks

from .utils import list_notebooks


def test_combine():
    nb_source = new_notebook(
        cells=[
            new_markdown_cell("Markdown text"),
            new_code_cell("a=3"),
            new_code_cell("a+1"),
            new_code_cell("a+1"),
            new_markdown_cell("Markdown text"),
            new_code_cell("a+2"),
        ]
    )

    nb_outputs = new_notebook(
        cells=[
            new_markdown_cell("Markdown text"),
            new_code_cell("a=3"),
            new_code_cell("a+1"),
            new_code_cell("a+2"),
            new_markdown_cell("Markdown text"),
        ]
    )

    nb_outputs.cells[2].outputs = ["4"]
    nb_outputs.cells[3].outputs = ["5"]

    nb_source = combine_inputs_with_outputs(nb_source, nb_outputs)

    assert nb_source.cells[2].outputs == ["4"]
    assert nb_source.cells[3].outputs == []
    assert nb_source.cells[5].outputs == ["5"]


def test_read_text_and_combine_with_outputs(tmpdir):
    tmp_ipynb = "notebook.ipynb"
    tmp_script = "notebook.py"

    with (open(str(tmpdir.join(tmp_script)), "w")) as fp:
        fp.write(
            """# ---
# jupyter:
#   jupytext_formats: ipynb,py:light
# ---

1+1

2+2

3+3
"""
        )

    with (open(str(tmpdir.join(tmp_ipynb)), "w")) as fp:
        fp.write(
            """{
 "cells": [
  {
   "cell_type": "code",
   "execution_count": 1,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "2"
      ]
     },
     "execution_count": 1,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "1+1"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 3,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "6"
      ]
     },
     "execution_count": 3,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "3+3"
   ]
  }
 ],
 "metadata": {},
 "nbformat": 4,
 "nbformat_minor": 2
}
"""
        )

    # create contents manager
    cm = jupytext.TextFileContentsManager()
    cm.root_dir = str(tmpdir)

    # load notebook from script
    model = cm.get(tmp_script)
    nb = model["content"]

    assert nb.cells[0]["source"] == "1+1"
    assert nb.cells[1]["source"] == "2+2"
    assert nb.cells[2]["source"] == "3+3"

    # No output for the second cell, which is not in the ipynb
    assert nb.cells[0]["outputs"]
    assert not nb.cells[1]["outputs"]
    assert nb.cells[2]["outputs"]

    assert len(nb.cells) == 3


@pytest.mark.parametrize("nb_file", list_notebooks("ipynb_all"))
def test_combine_stable(nb_file):
    nb_org = jupytext.read(nb_file)
    nb_source = deepcopy(nb_org)
    nb_outputs = deepcopy(nb_org)

    for cell in nb_source.cells:
        cell.outputs = []

    nb_source = combine_inputs_with_outputs(nb_source, nb_outputs)
    compare_notebooks(nb_source, nb_org)


def test_combine_reorder():
    nb_source = new_notebook(
        cells=[
            new_markdown_cell("Markdown text"),
            new_code_cell("1+1"),
            new_code_cell("2+2"),
            new_code_cell("3+3"),
            new_markdown_cell("Markdown text"),
            new_code_cell("4+4"),
        ]
    )

    nb_outputs = new_notebook(
        cells=[
            new_markdown_cell("Markdown text"),
            new_code_cell("2+2"),
            new_code_cell("4+4"),
            new_code_cell("1+1"),
            new_code_cell("3+3"),
            new_markdown_cell("Markdown text"),
        ]
    )

    nb_outputs.cells[1].outputs = ["4"]
    nb_outputs.cells[2].outputs = ["8"]
    nb_outputs.cells[3].outputs = ["2"]
    nb_outputs.cells[4].outputs = ["6"]

    nb_source = combine_inputs_with_outputs(nb_source, nb_outputs)

    assert nb_source.cells[1].outputs == ["2"]
    assert nb_source.cells[2].outputs == ["4"]
    assert nb_source.cells[3].outputs == ["6"]
    assert nb_source.cells[5].outputs == ["8"]


def test_combine_split():
    nb_source = new_notebook(cells=[new_code_cell("1+1"), new_code_cell("2+2")])

    nb_outputs = new_notebook(cells=[new_code_cell("1+1\n2+2")])

    nb_outputs.cells[0].outputs = ["4"]

    nb_source = combine_inputs_with_outputs(nb_source, nb_outputs)

    assert nb_source.cells[0].outputs == []
    assert nb_source.cells[1].outputs == ["4"]


def test_combine_refactor():
    nb_source = new_notebook(
        cells=[new_code_cell("a=1"), new_code_cell("a+1"), new_code_cell("a+2")]
    )

    nb_outputs = new_notebook(
        cells=[new_code_cell("b=1"), new_code_cell("b+1"), new_code_cell("b+2")]
    )

    nb_outputs.cells[1].outputs = ["2"]
    nb_outputs.cells[2].outputs = ["3"]

    nb_source = combine_inputs_with_outputs(nb_source, nb_outputs)

    assert nb_source.cells[0].outputs == []
    assert nb_source.cells[1].outputs == ["2"]
    assert nb_source.cells[2].outputs == ["3"]


def test_combine_attachments():
    nb_source = new_notebook(
        cells=[new_markdown_cell("![image.png](attachment:image.png)")]
    )

    nb_outputs = new_notebook(
        cells=[
            new_markdown_cell(
                "![image.png](attachment:image.png)",
                attachments={"image.png": {"image/png": "SOME_LONG_IMAGE_CODE...=="}},
            )
        ]
    )

    nb_source = combine_inputs_with_outputs(nb_source, nb_outputs)
    compare(nb_source, nb_outputs)



import nbformat
import pytest
from nbformat.v4.nbbase import new_code_cell, new_notebook

from jupytext.cli import jupytext
from jupytext.compare import compare
from jupytext.header import header_to_metadata_and_cell
from jupytext.jupytext import read, write


def test_pairing_through_config_leaves_ipynb_unmodified(tmpdir):
    cfg_file = tmpdir.join(".jupytext.yml")
    nb_file = tmpdir.join("notebook.ipynb")
    py_file = tmpdir.join("notebook.py")

    cfg_file.write("formats: 'ipynb,py'\n")
    nbformat.write(new_notebook(), str(nb_file))

    jupytext([str(nb_file), "--sync"])

    assert nb_file.isfile()
    assert py_file.isfile()

    nb = nbformat.read(nb_file, as_version=4)
    assert "jupytext" not in nb.metadata


def test_formats(tmpdir):
    tmpdir.join(".jupytext").write(
        '''# Default pairing
formats = "ipynb,py"'''
    )
    test = tmpdir.join("test.py")
    test.write("1 + 1\n")

    jupytext([str(test), "--sync"])

    assert tmpdir.join("test.ipynb").isfile()


def test_formats_with_suffix(tmpdir):
    tmpdir.join(".jupytext").write('formats = "ipynb,.nb.py"')
    test = tmpdir.join("test.py")
    test.write("1 + 1\n")

    test_nb = tmpdir.join("test.nb.py")
    test_nb.write("1 + 1\n")

    jupytext([str(test), "--sync"])
    assert not tmpdir.join("test.ipynb").isfile()

    jupytext([str(test_nb), "--sync"])
    assert tmpdir.join("test.ipynb").isfile()


def test_formats_does_not_apply_to_config_file(tmpdir):
    config = tmpdir.join(".jupytext.py")
    config.write('c.formats = "ipynb,py"')
    test = tmpdir.join("test.py")
    test.write("1 + 1\n")

    jupytext([str(test), str(config), "--sync"])

    assert tmpdir.join("test.ipynb").isfile()
    assert not tmpdir.join(".jupytext.ipynb").isfile()


def test_preferred_jupytext_formats_save(tmpdir):
    tmpdir.join(".jupytext.yml").write("preferred_jupytext_formats_save: jl:percent")
    tmp_ipynb = tmpdir.join("notebook.ipynb")
    tmp_jl = tmpdir.join("notebook.jl")

    nb = new_notebook(
        cells=[new_code_cell("1 + 1")], metadata={"jupytext": {"formats": "ipynb,jl"}}
    )

    write(nb, str(tmp_ipynb))
    jupytext([str(tmp_ipynb), "--sync"])

    with open(str(tmp_jl)) as stream:
        text_jl = stream.read()

    # Parse the YAML header
    metadata, _, _, _ = header_to_metadata_and_cell(text_jl.splitlines(), "#")
    assert metadata["jupytext"]["formats"] == "ipynb,jl:percent"


@pytest.mark.parametrize(
    "config",
    [
        # Way 1: preferred_jupytext_formats_save + formats
        """preferred_jupytext_formats_save: "python//py:percent"
formats: "ipynb,python//py"
""",
        # Way 2: formats
        "formats: ipynb,python//py:percent",
    ],
)
def test_save_using_preferred_and_default_format(config, tmpdir):
    tmpdir.join(".jupytext.yml").write(config)
    tmp_ipynb = tmpdir.join("notebook.ipynb")
    tmp_py = tmpdir.join("python").join("notebook.py")

    nb = new_notebook(cells=[new_code_cell("1 + 1")])

    write(nb, str(tmp_ipynb))
    jupytext([str(tmp_ipynb), "--sync"])

    # read py file
    nb_py = read(str(tmp_py))
    assert nb_py.metadata["jupytext"]["text_representation"]["format_name"] == "percent"


def test_hide_notebook_metadata(tmpdir, no_jupytext_version_number):
    tmpdir.join(".jupytext").write("hide_notebook_metadata = true")
    tmp_ipynb = tmpdir.join("notebook.ipynb")
    tmp_md = tmpdir.join("notebook.md")

    nb = new_notebook(
        cells=[new_code_cell("1 + 1")], metadata={"jupytext": {"formats": "ipynb,md"}}
    )

    write(nb, str(tmp_ipynb))
    jupytext([str(tmp_ipynb), "--sync"])

    with open(str(tmp_md)) as stream:
        text_md = stream.read()

    compare(
        text_md,
        """<!--

---
jupyter:
  jupytext:
    formats: ipynb,md
    hide_notebook_metadata: true
---

-->

```python
1 + 1
```
""",
    )


def test_cli_config_on_windows_issue_629(tmpdir):
    cfg_file = tmpdir.join("jupytext.yml")
    cfg_file.write(
        """formats: "notebooks///ipynb,scripts///py:percent"
notebook_metadata_filter: "jupytext"
"""
    )

    tmpdir.mkdir("scripts").join("test.py").write("# %%\n 1+1\n")

    jupytext(["--sync", str(tmpdir.join("scripts").join("*.py"))])

    assert tmpdir.join("notebooks").join("test.ipynb").exists()


def test_sync_config_does_not_create_formats_metadata(
    tmpdir, cwd_tmpdir, python_notebook
):
    tmpdir.join("jupytext.yml").write(
        """formats: "ipynb,py:percent"
"""
    )

    write(python_notebook, "test.ipynb")
    jupytext(["--sync", "test.ipynb"])

    nb = read("test.py")
    assert "formats" not in nb.metadata["jupytext"]


def test_multiple_formats_771(tmpdir, cwd_tmpdir, python_notebook):
    tmpdir.join("jupytext.toml").write(
        """formats = "notebooks///ipynb,notebooks///py,scripts///py:percent"
"""
    )
    notebooks_dir = tmpdir.mkdir("notebooks")
    scripts_dir = tmpdir.join("scripts")

    write(python_notebook, str(notebooks_dir.join("notebook.ipynb")))
    jupytext(["--sync", "notebooks/notebook.ipynb"])

    assert notebooks_dir.join("notebook.py").isfile()
    assert scripts_dir.join("notebook.py").isfile()

    notebooks_dir.join("module.py").write("1 + 1\n")
    jupytext(["--sync", "notebooks/module.py"])

    assert notebooks_dir.join("module.ipynb").isfile()
    assert scripts_dir.join("module.py").isfile()



import re
import sys
from pathlib import Path

import pytest


def replace_issue_number_with_links(text):
    return re.sub(
        r"([^\[])#([0-9]+)",
        r"\1[#\2](https://github.com/mwouts/jupytext/issues/\2)",
        text,
    )


@pytest.mark.parametrize(
    "input,output",
    [
        (
            "Issue #535",
            "Issue [#535](https://github.com/mwouts/jupytext/issues/535)",
        ),
        (
            "Multiline\ntext (#123)",
            "Multiline\ntext ([#123](https://github.com/mwouts/jupytext/issues/123))",
        ),
    ],
)
def test_replace_issue_numbers_with_links(input, output):
    assert replace_issue_number_with_links(input) == output


@pytest.mark.skipif(
    sys.version_info < (3, 5), reason="'PosixPath' object has no attribute 'read_text'"
)
def test_update_changelog():
    changelog_file = Path(__file__).parent.parent / "docs" / "CHANGELOG.md"
    cur_text = changelog_file.read_text()
    new_text = replace_issue_number_with_links(cur_text)
    if cur_text != new_text:
        changelog_file.write_text(new_text)  # pragma: no cover



# -*- coding: utf-8 -*-

import jupytext
from jupytext.compare import compare


def test_read_simple_file(
    script="""# ---
# title: Simple file
# ---

# %% [markdown]
# This is a markdown cell

# %% [raw]
# This is a raw cell

# %%% sub-cell title
# This is a sub-cell

# %%%% sub-sub-cell title
# This is a sub-sub-cell

# %% And now a code cell
1 + 2 + 3 + 4
5
6
%%pylab inline

7
""",
):
    nb = jupytext.reads(script, "py:hydrogen")
    assert len(nb.cells) == 6
    assert nb.cells[0].cell_type == "raw"
    assert nb.cells[0].source == "---\ntitle: Simple file\n---"
    assert nb.cells[1].cell_type == "markdown"
    assert nb.cells[1].source == "This is a markdown cell"
    assert nb.cells[2].cell_type == "raw"
    assert nb.cells[2].source == "This is a raw cell"
    assert nb.cells[3].cell_type == "code"
    assert nb.cells[3].source == "# This is a sub-cell"
    assert nb.cells[3].metadata["title"] == "sub-cell title"
    assert nb.cells[4].cell_type == "code"
    assert nb.cells[4].source == "# This is a sub-sub-cell"
    assert nb.cells[4].metadata["title"] == "sub-sub-cell title"
    assert nb.cells[5].cell_type == "code"
    compare(
        nb.cells[5].source,
        """1 + 2 + 3 + 4
5
6
%%pylab inline

7""",
    )
    assert nb.cells[5].metadata == {"title": "And now a code cell"}

    script2 = jupytext.writes(nb, "py:hydrogen")
    compare(script2, script)


def test_read_cell_with_metadata(
    script="""# %% a code cell with parameters {"tags": ["parameters"]}
a = 3
""",
):
    nb = jupytext.reads(script, "py:hydrogen")
    assert len(nb.cells) == 1
    assert nb.cells[0].cell_type == "code"
    assert nb.cells[0].source == "a = 3"
    assert nb.cells[0].metadata == {
        "title": "a code cell with parameters",
        "tags": ["parameters"],
    }

    script2 = jupytext.writes(nb, "py:hydrogen")
    compare(script2, script)


def test_read_nbconvert_script(
    script="""
# coding: utf-8

# A markdown cell

# In[1]:


%pylab inline
import pandas as pd

pd.options.display.max_rows = 6
pd.options.display.max_columns = 20


# Another markdown cell

# In[2]:


1 + 1


# Again, a markdown cell

# In[33]:


2 + 2


# <codecell>


3 + 3
""",
):
    assert jupytext.formats.guess_format(script, ".py")[0] == "hydrogen"
    nb = jupytext.reads(script, ".py")
    assert len(nb.cells) == 5


def test_read_remove_blank_lines(
    script="""# %%
import pandas as pd

# %% Display a data frame
df = pd.DataFrame({'A': [1, 2], 'B': [3, 4]},
                  index=pd.Index(['x0', 'x1'], name='x'))
df

# %% Pandas plot {"tags": ["parameters"]}
df.plot(kind='bar')


# %% sample class
class MyClass:
    pass


# %% a function
def f(x):
    return 42 * x

""",
):
    nb = jupytext.reads(script, "py")
    assert len(nb.cells) == 5
    for i in range(5):
        assert nb.cells[i].cell_type == "code"
        assert not nb.cells[i].source.startswith("\n")
        assert not nb.cells[i].source.endswith("\n")

    script2 = jupytext.writes(nb, "py:hydrogen")
    compare(script2, script)


def test_no_crash_on_square_bracket(
    script="""# %% In [2]
print('Hello')
""",
):
    nb = jupytext.reads(script, "py")
    script2 = jupytext.writes(nb, "py:hydrogen")
    compare(script2, script)


def test_nbconvert_cell(
    script="""# In[2]:
print('Hello')
""",
):
    nb = jupytext.reads(script, "py")
    script2 = jupytext.writes(nb, "py:hydrogen")
    expected = """# %%
print('Hello')
"""
    compare(script2, expected)


def test_nbformat_v3_nbpy_cell(
    script="""# <codecell>
print('Hello')
""",
):
    nb = jupytext.reads(script, "py")
    script2 = jupytext.writes(nb, "py:hydrogen")
    expected = """# %%
print('Hello')
"""
    compare(script2, expected)



import pytest

import jupytext
from jupytext.compare import compare


@pytest.mark.parametrize("lang", ["cs", "c#", "csharp"])
def test_simple_cs(lang):
    source = """// A Hello World! program in C#.
Console.WriteLine("Hello World!");
"""
    md = """```{lang}
{source}
```
""".format(
        lang=lang, source=source
    )
    nb = jupytext.reads(md, "md")
    assert nb.metadata["jupytext"]["main_language"] == "csharp"
    assert len(nb.cells) == 1
    assert nb.cells[0].cell_type == "code"

    cs = jupytext.writes(nb, "cs")
    assert source in cs
    if lang != "csharp":
        assert cs.startswith('// + language="{lang}"'.format(lang=lang))

    md2 = jupytext.writes(nb, "md")
    compare(md2, md)


@pytest.mark.parametrize("lang", ["cs", "c#", "csharp"])
def test_csharp_magics(no_jupytext_version_number, lang):
    md = """```{lang}
#!html
<b>Hello!</b>
```
""".format(
        lang=lang
    )
    nb = jupytext.reads(md, "md")
    nb.metadata["jupytext"].pop("notebook_metadata_filter")
    nb.metadata["jupytext"].pop("cell_metadata_filter")
    assert nb.metadata["jupytext"]["main_language"] == "csharp"
    assert len(nb.cells) == 1
    assert nb.cells[0].cell_type == "code"

    cs = jupytext.writes(nb, "cs")
    assert all(line.startswith("//") for line in cs.splitlines()), cs

    md2 = jupytext.writes(nb, "md")
    md_expected = """---
jupyter:
  jupytext:
    main_language: csharp
---

```html
<b>Hello!</b>
```
"""
    compare(md2, md_expected)


def test_read_html_cell_from_md(no_jupytext_version_number):
    md = """---
jupyter:
  jupytext:
    main_language: csharp
---

```html
<b>Hello!</b>
```
"""

    nb = jupytext.reads(md, "md")
    assert len(nb.cells) == 1
    assert nb.cells[0].cell_type == "code"
    compare(nb.cells[0].source, "#!html\n<b>Hello!</b>")

    md2 = jupytext.writes(nb, "md")
    compare(md2, md)



"""Combine source and outputs from two notebooks
"""
import re
from copy import copy

from nbformat import NotebookNode

from .cell_metadata import _IGNORE_CELL_METADATA
from .formats import long_form_one_format
from .header import _DEFAULT_NOTEBOOK_METADATA
from .metadata_filter import restore_filtered_metadata

_BLANK_LINE = re.compile(r"^\s*$")


def black_invariant(text, chars=None):
    """Remove characters that may be changed when reformatting the text with black"""
    if chars is None:
        chars = [" ", "\t", "\n", ",", "'", '"', "(", ")", "\\"]

    for char in chars:
        text = text.replace(char, "")
    return text


def same_content(ref, test, endswith=False):
    """Is the content of two cells the same, up to reformatting by black"""
    ref = black_invariant(ref)
    test = black_invariant(test)

    if endswith and test:
        return ref.endswith(test)
    return ref == test


def combine_inputs_with_outputs(nb_source, nb_outputs, fmt=None):
    """Return a notebook that combines the text and metadata from the first notebook,
    with the outputs and metadata of the second notebook."""
    # nbformat version number taken from the notebook with outputs
    assert (
        nb_outputs.nbformat == nb_source.nbformat
    ), "The notebook with outputs is in format {}.{}, please upgrade it to {}.x".format(
        nb_outputs.nbformat, nb_outputs.nbformat_minor, nb_source.nbformat
    )
    nb_source.nbformat_minor = nb_outputs.nbformat_minor

    fmt = long_form_one_format(fmt)
    text_repr = nb_source.metadata.get("jupytext", {}).get("text_representation", {})
    ext = fmt.get("extension") or text_repr.get("extension")
    format_name = fmt.get("format_name") or text_repr.get("format_name")

    notebook_metadata_filter = nb_source.metadata.get("jupytext", {}).get(
        "notebook_metadata_filter"
    )
    if notebook_metadata_filter == "-all":
        nb_metadata = nb_outputs.metadata

    else:
        nb_metadata = restore_filtered_metadata(
            nb_source.metadata,
            nb_outputs.metadata,
            notebook_metadata_filter,
            _DEFAULT_NOTEBOOK_METADATA,
        )

    source_is_md_version_one = (
        ext in [".md", ".markdown", ".Rmd"] and text_repr.get("format_version") == "1.0"
    )
    if nb_metadata.get("jupytext", {}).get("formats") or ext in [
        ".md",
        ".markdown",
        ".Rmd",
    ]:
        nb_metadata.get("jupytext", {}).pop("text_representation", None)

    if not nb_metadata.get("jupytext", {}):
        nb_metadata.pop("jupytext", {})

    if format_name in ["nomarker", "sphinx"] or source_is_md_version_one:
        cell_metadata_filter = "-all"
    else:
        cell_metadata_filter = nb_metadata.get("jupytext", {}).get(
            "cell_metadata_filter"
        )

    outputs_map = map_outputs_to_inputs(nb_source.cells, nb_outputs.cells)

    cells = []
    for source_cell, j in zip(nb_source.cells, outputs_map):
        if j is None:
            cells.append(source_cell)
            continue

        output_cell = nb_outputs.cells[j]

        # Outputs and optional attributes are taken from the notebook with outputs
        cell = copy(output_cell)

        # Cell text is taken from the source notebook
        cell.source = source_cell.source

        # We also restore the cell metadata that has been filtered
        cell.metadata = restore_filtered_metadata(
            source_cell.metadata,
            output_cell.metadata,
            # The 'spin' format does not allow metadata on non-code cells
            "-all"
            if format_name == "spin" and source_cell.cell_type != "code"
            else cell_metadata_filter,
            _IGNORE_CELL_METADATA,
        )

        cells.append(cell)

    # We call NotebookNode rather than new_notebook as we don't want to validate
    # the notebook (some of the notebook in the collection of test notebooks
    # do have some invalid properties - probably inherited from an older version
    # of the notebook format).
    return NotebookNode(
        cells=cells,
        metadata=nb_metadata,
        nbformat=nb_outputs.nbformat,
        nbformat_minor=nb_outputs.nbformat_minor,
    )


def map_outputs_to_inputs(cells_inputs, cells_outputs):
    """Returns a map i->(j or None) that maps the cells with outputs to the input cells"""
    n_in = len(cells_inputs)
    n_out = len(cells_outputs)
    outputs_map = [None] * n_in

    # First rule: match based on cell type, content, in increasing order, for each cell type
    first_unmatched_output_per_cell_type = {}
    for i in range(n_in):
        cell_input = cells_inputs[i]
        for j in range(
            first_unmatched_output_per_cell_type.get(cell_input.cell_type, 0), n_out
        ):
            cell_output = cells_outputs[j]
            if cell_input.cell_type == cell_output.cell_type and same_content(
                cell_input.source, cell_output.source
            ):
                outputs_map[i] = j
                first_unmatched_output_per_cell_type[cell_input.cell_type] = j + 1
                break

    # Second rule: match unused outputs based on cell type and content
    # Third rule: is the new cell the final part of a previous cell with outputs?
    unused_ouputs = set(range(n_out)).difference(outputs_map)
    for endswith in [False, True]:
        if not unused_ouputs:
            return outputs_map

        for i in range(n_in):
            if outputs_map[i] is not None:
                continue
            cell_input = cells_inputs[i]
            for j in unused_ouputs:
                cell_output = cells_outputs[j]
                if cell_input.cell_type == cell_output.cell_type and same_content(
                    cell_output.source, cell_input.source, endswith
                ):
                    outputs_map[i] = j
                    unused_ouputs.remove(j)
                    break

    # Fourth rule: match based on increasing index (and cell type) for non-empty cells
    if not unused_ouputs:
        return outputs_map

    prev_j = -1
    for i in range(n_in):
        if outputs_map[i] is not None:
            prev_j = outputs_map[i]
            continue

        j = prev_j + 1
        if j not in unused_ouputs:
            continue

        cell_input = cells_inputs[i]
        cell_output = cells_outputs[j]
        if (
            cell_input.cell_type == cell_output.cell_type
            and cell_input.source.strip() != ""
        ):
            outputs_map[i] = j
            unused_ouputs.remove(j)
            prev_j = j

    return outputs_map



import pytest
from git.exc import HookExecutionError
from nbformat.v4.nbbase import new_markdown_cell, new_notebook
from pre_commit.main import main as pre_commit

from jupytext import read, write
from jupytext.cli import jupytext
from jupytext.compare import compare_cells

from .utils import (
    skip_pre_commit_tests_on_windows,
    skip_pre_commit_tests_when_jupytext_folder_is_not_a_git_repo,
)


@skip_pre_commit_tests_on_windows
@skip_pre_commit_tests_when_jupytext_folder_is_not_a_git_repo
def test_pre_commit_hook_ipynb_to_py(
    tmpdir, cwd_tmpdir, tmp_repo, jupytext_repo_root, jupytext_repo_rev
):
    """Here we document and test the expected behavior of the pre-commit hook in the
    directional (--to) mode. Note that here, the ipynb file is always the source for
    updates - i.e. changes on the .py file will not trigger the hook.
    """
    # set up the tmpdir repo with pre-commit
    pre_commit_config_yaml = f"""
repos:
- repo: {jupytext_repo_root}
  rev: {jupytext_repo_rev}
  hooks:
  - id: jupytext
    args: [--from, ipynb, --to, "py:percent"]
"""

    tmpdir.join(".pre-commit-config.yaml").write(pre_commit_config_yaml)
    tmp_repo.git.add(".pre-commit-config.yaml")
    pre_commit(["install", "--install-hooks"])

    # write test notebook and output file
    nb = new_notebook(cells=[new_markdown_cell("A short notebook")])
    write(nb, "test.ipynb")
    jupytext(["--from", "ipynb", "--to", "py:percent", "test.ipynb"])

    tmp_repo.git.add(".")
    tmp_repo.index.commit("test")

    # make a change to the notebook
    nb = new_notebook(cells=[new_markdown_cell("Some other text")])
    write(nb, "test.ipynb")

    tmp_repo.git.add("test.ipynb")
    # now a commit will fail, and keep failing until we add the new
    # changes made to the existing output to the index ourselves
    with pytest.raises(HookExecutionError, match="files were modified by this hook"):
        tmp_repo.index.commit("fails")

    with pytest.raises(HookExecutionError, match="git add test.py"):
        tmp_repo.index.commit("fails again")

    # once we add the changes, it will pass
    tmp_repo.git.add("test.py")
    tmp_repo.index.commit("succeeds")

    assert "test.ipynb" in tmp_repo.tree()
    assert "test.py" in tmp_repo.tree()

    # Updating the .py file is possible
    nb = new_notebook(cells=[new_markdown_cell("Some updated text")])
    write(nb, "test.py", fmt="py:percent")
    tmp_repo.index.commit("update py version")

    # But it won't change the ipynb file (if you want that, use the --sync mode)
    nb = read("test.ipynb")
    compare_cells(nb.cells, [new_markdown_cell("Some other text")], compare_ids=False)



from copy import deepcopy
from itertools import product

import pytest

from jupytext import read, write, writes
from jupytext.compare import compare
from jupytext.formats import long_form_one_format

from .utils import list_notebooks


@pytest.mark.parametrize(
    "nb_file,fmt",
    product(
        list_notebooks("ipynb_py") + list_notebooks("ipynb_R"),
        ["auto:light", "auto:percent", "md", ".Rmd", ".ipynb"],
    ),
)
def test_write_notebook_does_not_change_it(nb_file, fmt, tmpdir):
    nb_org = read(nb_file)
    nb_org_copied = deepcopy(nb_org)
    ext = long_form_one_format(fmt, nb_org.metadata)["extension"]

    writes(nb_org, fmt)
    compare(nb_org, nb_org_copied)

    tmp_dest = str(tmpdir.join("notebook" + ext))
    write(nb_org, tmp_dest, fmt=fmt)
    compare(nb_org, nb_org_copied)



# -*- coding: utf-8 -*-

import jupytext
from jupytext.compare import compare


def test_read_simple_file(
    script="""// ---
// title: Simple file
// ---

// Here we have some text
// And below we have some code

System.out.println("Hello World");
""",
):
    nb = jupytext.reads(script, "java")
    assert len(nb.cells) == 3
    assert nb.cells[0].cell_type == "raw"
    assert nb.cells[0].source == "---\ntitle: Simple file\n---"
    assert nb.cells[1].cell_type == "markdown"
    assert (
        nb.cells[1].source == "Here we have some text\n" "And below we have some code"
    )
    assert nb.cells[2].cell_type == "code"
    assert nb.cells[2].source == """System.out.println("Hello World");"""

    script2 = jupytext.writes(nb, "java")
    compare(script2, script)



import nbformat
import pytest

import jupytext
from jupytext.cli import jupytext as jupytext_cli
from jupytext.compare import compare, compare_notebooks
from jupytext.doxygen import doxygen_to_markdown, markdown_to_doxygen

SAMPLE_MARKDOWN = """A latex formula $$1+1$$, another one $$2+2
$$

An inline formula $3+3$
"""

SAMPLE_DOXYGEN = """A latex formula \\f[1+1\\f], another one \\f[2+2
\\f]

An inline formula \\f$3+3\\f$
"""


def test_markdown_to_doxygen():
    compare(markdown_to_doxygen(SAMPLE_MARKDOWN), SAMPLE_DOXYGEN)


def test_doxygen_to_markdown():
    compare(doxygen_to_markdown(SAMPLE_DOXYGEN), SAMPLE_MARKDOWN)


@pytest.mark.parametrize(
    "latex,doxygen",
    [
        ("$$", "$$"),
        ("$1+1$", "\\f$1+1\\f$"),
        ("$1\\$ $", "\\f$1\\$ \\f$"),
        ("$$2+2$$", "\\f[2+2\\f]"),
        ("$$2+2$$ then $$3+3$$", "\\f[2+2\\f] then \\f[3+3\\f]"),
        ("$$2+2\n$$", "\\f[2+2\n\\f]"),
        ("$$1\\$+2\\$=3\\$$$", "\\f[1\\$+2\\$=3\\$\\f]"),
    ],
)
def test_simple_equations_to_doxygen_and_back(latex, doxygen):
    assert markdown_to_doxygen(latex) == doxygen
    assert doxygen_to_markdown(doxygen) == latex


def test_doxygen_equation_markers(tmpdir):
    cfg_file = tmpdir.join("jupytext.toml")
    nb_file = tmpdir.join("notebook.ipynb")
    md_file = tmpdir.join("notebook.md")

    cfg_file.write("hide_notebook_metadata = true\ndoxygen_equation_markers = true")
    nb = jupytext.reads(SAMPLE_MARKDOWN, "md")
    del nb.metadata["jupytext"]
    nbformat.write(nb, str(nb_file))

    jupytext_cli([str(nb_file), "--to", "md"])
    md = md_file.read()

    # Doxygen equation markers
    assert "$$" not in md
    assert "\\f[" in md
    assert "\\f]" in md
    assert "\\f$" in md
    assert "\\f$" in md

    # Metadata hidden
    assert md.startswith("<!--\n\n---\n")

    jupytext_cli([str(md_file), "--to", "notebook"])
    nb2 = nbformat.read(str(nb_file), as_version=4)

    compare_notebooks(nb2, nb)



import os
import shutil

import pytest

from jupytext.compare import compare_notebooks
from jupytext.contentsmanager import TextFileContentsManager

from .utils import list_notebooks


@pytest.mark.parametrize("nb_file", list_notebooks("python"))
def test_py_notebooks_are_trusted(nb_file):
    cm = TextFileContentsManager()
    root, file = os.path.split(nb_file)
    cm.root_dir = root
    nb = cm.get(file)
    for cell in nb["content"].cells:
        assert cell.metadata.get("trusted", True)


@pytest.mark.parametrize("nb_file", list_notebooks("Rmd"))
def test_rmd_notebooks_are_trusted(nb_file):
    cm = TextFileContentsManager()
    root, file = os.path.split(nb_file)
    cm.root_dir = root
    nb = cm.get(file)
    for cell in nb["content"].cells:
        assert cell.metadata.get("trusted", True)


@pytest.mark.parametrize("nb_file", list_notebooks("ipynb_py", skip="hash sign"))
def test_ipynb_notebooks_can_be_trusted(nb_file, tmpdir, no_jupytext_version_number):
    cm = TextFileContentsManager()
    root, file = os.path.split(nb_file)
    tmp_ipynb = str(tmpdir.join(file))
    py_file = file.replace(".ipynb", ".py")
    tmp_py = str(tmpdir.join(py_file))
    shutil.copy(nb_file, tmp_ipynb)

    cm.formats = "ipynb,py"
    cm.root_dir = str(tmpdir)
    model = cm.get(file)
    cm.save(model, py_file)

    # Unsign and test notebook
    nb = model["content"]
    for cell in nb.cells:
        cell.metadata.pop("trusted", True)

    cm.notary.unsign(nb)

    model = cm.get(file)
    for cell in model["content"].cells:
        assert (
            "trusted" not in cell.metadata
            or not cell.metadata["trusted"]
            or not cell.outputs
        )

    # Trust and reload
    cm.trust_notebook(py_file)

    model = cm.get(file)
    for cell in model["content"].cells:
        assert cell.metadata.get("trusted", True)

    # Remove py file, content should be the same
    os.remove(tmp_py)
    nb2 = cm.get(file)
    for cell in nb2["content"].cells:
        assert cell.metadata.get("trusted", True)

    compare_notebooks(nb2["content"], model["content"])

    # Just for coverage
    cm.trust_notebook(file)


@pytest.mark.parametrize("nb_file", list_notebooks("ipynb_py", skip="hash sign"))
def test_ipynb_notebooks_can_be_trusted_even_with_metadata_filter(
    nb_file, tmpdir, no_jupytext_version_number
):
    cm = TextFileContentsManager()
    root, file = os.path.split(nb_file)
    tmp_ipynb = str(tmpdir.join(file))
    py_file = file.replace(".ipynb", ".py")
    tmp_py = str(tmpdir.join(py_file))
    shutil.copy(nb_file, tmp_ipynb)

    cm.formats = "ipynb,py"
    cm.notebook_metadata_filter = "all"
    cm.cell_metadata_filter = "-all"
    cm.root_dir = str(tmpdir)
    model = cm.get(file)
    cm.save(model, py_file)

    # Unsign notebook
    nb = model["content"]
    for cell in nb.cells:
        cell.metadata.pop("trusted", True)

    cm.notary.unsign(nb)

    # Trust and reload
    cm.trust_notebook(py_file)

    model = cm.get(file)
    for cell in model["content"].cells:
        assert cell.metadata.get("trusted", True)

    # Remove py file, content should be the same
    os.remove(tmp_py)
    nb2 = cm.get(file)
    for cell in nb2["content"].cells:
        assert cell.metadata.get("trusted", True)

    compare_notebooks(nb2["content"], model["content"])


@pytest.mark.parametrize("nb_file", list_notebooks("percent", skip="hash sign"))
def test_text_notebooks_can_be_trusted(nb_file, tmpdir, no_jupytext_version_number):
    cm = TextFileContentsManager()
    root, file = os.path.split(nb_file)
    py_file = str(tmpdir.join(file))
    shutil.copy(nb_file, py_file)

    cm.root_dir = str(tmpdir)
    model = cm.get(file)
    model["type"] == "notebook"
    cm.save(model, file)

    # Unsign notebook
    nb = model["content"]
    for cell in nb.cells:
        cell.metadata.pop("trusted", True)

    cm.notary.unsign(nb)

    # Trust and reload
    cm.trust_notebook(file)

    model = cm.get(file)
    for cell in model["content"].cells:
        assert cell.metadata.get("trusted", True)


def test_simple_notebook_is_trusted(tmpdir, python_notebook):
    cm = TextFileContentsManager()
    cm.root_dir = str(tmpdir)

    nb = python_notebook
    cm.notary.unsign(nb)

    # All cells are trusted in this notebook
    assert cm.notary.check_cells(nb)
    # Yet the notebook is not in the database of trusted notebooks
    assert not cm.notary.check_signature(nb)

    # Save the notebook using the CM
    cm.save(dict(type="notebook", content=nb), "test.ipynb")

    # The notebook is safe so it should have been trusted before getting saved to disk
    nb = cm.get("test.ipynb")["content"]
    assert cm.notary.check_signature(nb)



from nbformat.v4.nbbase import new_code_cell


def test_cell_id_is_not_random():
    id1 = new_code_cell().id
    id2 = new_code_cell().id

    n1 = int(id1.split("-")[1])
    n2 = int(id2.split("-")[1])
    assert n2 == n1 + 1



import pytest
from nbformat import NotebookNode

import jupytext
from jupytext.compare import compare, compare_cells

HEADER = {
    ".py": """# ---
# jupyter:
#   jupytext:
#     main_language: python
# ---

""",
    ".R": """# ---
# jupyter:
#   jupytext:
#     main_language: python
# ---

""",
    ".md": """---
jupyter:
  jupytext:
    main_language: python
---

""",
    ".Rmd": """---
jupyter:
  jupytext:
    main_language: python
---

""",
}

ACTIVE_ALL = {
    ".py": """# + active="ipynb,py,R,Rmd"
# This cell is active in all extensions
""",
    ".Rmd": """```{python active="ipynb,py,R,Rmd"}
# This cell is active in all extensions
```
""",
    ".md": """```python active="ipynb,py,R,Rmd"
# This cell is active in all extensions
```
""",
    ".R": """# + active="ipynb,py,R,Rmd"
# This cell is active in all extensions
""",
    ".ipynb": {
        "cell_type": "code",
        "source": "# This cell is active in all extensions",
        "metadata": {"active": "ipynb,py,R,Rmd"},
        "execution_count": None,
        "outputs": [],
    },
}


def check_active_cell(ext, active_dict):
    text = ("" if ext == ".py" else HEADER[ext]) + active_dict[ext]
    nb = jupytext.reads(text, ext)
    assert len(nb.cells) == 1
    compare(jupytext.writes(nb, ext), text)
    cell = NotebookNode(active_dict[".ipynb"])
    compare_cells(nb.cells, [cell], compare_ids=False)


@pytest.mark.parametrize("ext", [".Rmd", ".md", ".py", ".R"])
def test_active_all(ext, no_jupytext_version_number):
    check_active_cell(ext, ACTIVE_ALL)


ACTIVE_IPYNB = {
    ".py": """# + active="ipynb"
# # This cell is active only in ipynb
# %matplotlib inline
""",
    ".Rmd": """```{python active="ipynb", eval=FALSE}
# This cell is active only in ipynb
%matplotlib inline
```
""",
    ".md": """```python active="ipynb"
# This cell is active only in ipynb
%matplotlib inline
```
""",
    ".R": """# + active="ipynb"
# # This cell is active only in ipynb
# %matplotlib inline
""",
    ".ipynb": {
        "cell_type": "code",
        "source": "# This cell is active only in ipynb\n" "%matplotlib inline",
        "metadata": {"active": "ipynb"},
        "execution_count": None,
        "outputs": [],
    },
}


@pytest.mark.parametrize("ext", [".Rmd", ".md", ".py", ".R"])
def test_active_ipynb(ext, no_jupytext_version_number):
    check_active_cell(ext, ACTIVE_IPYNB)


ACTIVE_IPYNB_RMD_USING_TAG = {
    ".py": """# + tags=["active-ipynb-Rmd"]
# # This cell is active only in ipynb and Rmd
# %matplotlib inline
""",
    ".Rmd": """```{python tags=c("active-ipynb-Rmd")}
# This cell is active only in ipynb and Rmd
# %matplotlib inline
```
""",
    ".md": """```python tags=["active-ipynb-Rmd"]
# This cell is active only in ipynb and Rmd
%matplotlib inline
```
""",
    ".R": """# + tags=["active-ipynb-Rmd"]
# # This cell is active only in ipynb and Rmd
# %matplotlib inline
""",
    ".ipynb": {
        "cell_type": "code",
        "source": "# This cell is active only in ipynb and Rmd\n" "%matplotlib inline",
        "metadata": {"tags": ["active-ipynb-Rmd"]},
        "execution_count": None,
        "outputs": [],
    },
}


@pytest.mark.parametrize("ext", [".Rmd", ".md", ".py", ".R"])
def test_active_ipynb_rmd_using_tags(ext, no_jupytext_version_number):
    check_active_cell(ext, ACTIVE_IPYNB_RMD_USING_TAG)


ACTIVE_IPYNB_RSPIN = {
    ".R": """#+ active="ipynb", eval=FALSE
# # This cell is active only in ipynb
# 1 + 1
""",
    ".ipynb": {
        "cell_type": "code",
        "source": "# This cell is active only in ipynb\n" "1 + 1",
        "metadata": {"active": "ipynb"},
        "execution_count": None,
        "outputs": [],
    },
}


def test_active_ipynb_rspin(no_jupytext_version_number):
    nb = jupytext.reads(ACTIVE_IPYNB_RSPIN[".R"], "R:spin")
    assert len(nb.cells) == 1
    compare(jupytext.writes(nb, "R:spin"), ACTIVE_IPYNB_RSPIN[".R"])
    cell = NotebookNode(ACTIVE_IPYNB_RSPIN[".ipynb"])
    compare_cells(nb.cells, [cell], compare_ids=False)


ACTIVE_PY_IPYNB = {
    ".py": """# + active="ipynb,py"
# This cell is active in py and ipynb extensions
""",
    ".Rmd": """```{python active="ipynb,py", eval=FALSE}
# This cell is active in py and ipynb extensions
```
""",
    ".md": """```python active="ipynb,py"
# This cell is active in py and ipynb extensions
```
""",
    ".R": """# + active="ipynb,py"
# # This cell is active in py and ipynb extensions
""",
    ".ipynb": {
        "cell_type": "code",
        "source": "# This cell is active in py and " "ipynb extensions",
        "metadata": {"active": "ipynb,py"},
        "execution_count": None,
        "outputs": [],
    },
}


@pytest.mark.parametrize("ext", [".Rmd", ".md", ".py", ".R"])
def test_active_py_ipynb(ext, no_jupytext_version_number):
    check_active_cell(ext, ACTIVE_PY_IPYNB)


ACTIVE_PY_R_IPYNB = {
    ".py": """# + active="ipynb,py,R"
# This cell is active in py, R and ipynb extensions
""",
    ".Rmd": """```{python active="ipynb,py,R", eval=FALSE}
# This cell is active in py, R and ipynb extensions
```
""",
    ".R": """# + active="ipynb,py,R"
# This cell is active in py, R and ipynb extensions
""",
    ".ipynb": {
        "cell_type": "code",
        "source": "# This cell is active in py, R and " "ipynb extensions",
        "metadata": {"active": "ipynb,py,R"},
        "execution_count": None,
        "outputs": [],
    },
}


@pytest.mark.parametrize("ext", [".Rmd", ".py", ".R"])
def test_active_py_r_ipynb(ext, no_jupytext_version_number):
    check_active_cell(ext, ACTIVE_PY_R_IPYNB)


ACTIVE_RMD = {
    ".py": """# + active="Rmd"
# # This cell is active in Rmd only
""",
    ".Rmd": """```{python active="Rmd"}
# This cell is active in Rmd only
```
""",
    ".R": """# + active="Rmd"
# # This cell is active in Rmd only
""",
    ".ipynb": {
        "cell_type": "raw",
        "source": "# This cell is active in Rmd only",
        "metadata": {"active": "Rmd"},
    },
}


@pytest.mark.parametrize("ext", [".Rmd", ".py", ".R"])
def test_active_rmd(ext, no_jupytext_version_number):
    check_active_cell(ext, ACTIVE_RMD)


ACTIVE_NOT_INCLUDE_RMD = {
    ".py": """# + tags=["remove_cell"] active="Rmd"
# # This cell is active in Rmd only
""",
    ".Rmd": """```{python include=FALSE, active="Rmd"}
# This cell is active in Rmd only
```
""",
    ".R": """# + tags=["remove_cell"] active="Rmd"
# # This cell is active in Rmd only
""",
    ".ipynb": {
        "cell_type": "raw",
        "source": "# This cell is active in Rmd only",
        "metadata": {"active": "Rmd", "tags": ["remove_cell"]},
    },
}


@pytest.mark.parametrize("ext", [".Rmd", ".py", ".R"])
def test_active_not_include_rmd(ext, no_jupytext_version_number):
    check_active_cell(ext, ACTIVE_NOT_INCLUDE_RMD)


def test_active_cells_from_py_percent(
    text="""# %% active="py"
print('should only be displayed in py file')

# %% tags=["active-py"]
print('should only be displayed in py file')

# %% active="ipynb"
# print('only in jupyter')
""",
):
    """Example taken from https://github.com/mwouts/jupytext/issues/477"""
    nb = jupytext.reads(text, "py:percent")
    assert nb.cells[0].cell_type == "raw"
    assert nb.cells[1].cell_type == "raw"
    assert nb.cells[2].cell_type == "code"
    assert nb.cells[2].source == "print('only in jupyter')"

    text2 = jupytext.writes(nb, "py:percent")
    compare(text2, text)


def test_active_cells_from_py_light(
    text="""# + active="py"
print('should only be displayed in py file')

# + tags=["active-py"]
print('should only be displayed in py file')

# + active="ipynb"
# print('only in jupyter')
""",
):
    """Example adapted from https://github.com/mwouts/jupytext/issues/477"""
    nb = jupytext.reads(text, "py")
    assert nb.cells[0].cell_type == "raw"
    assert nb.cells[1].cell_type == "raw"
    assert nb.cells[2].cell_type == "code"
    assert nb.cells[2].source == "print('only in jupyter')"

    text2 = jupytext.writes(nb, "py")
    compare(text2, text)
